"""Tigramite causal discovery for time series."""

# Authors: Wiebke Günther <wiebke.guenther@dlr.de>, Urmi Ninad, Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

from __future__ import print_function
import numpy as np
from tigramite.pcmci import PCMCI
from copy import deepcopy
import itertools

from tigramite.toymodels.context_model import _group_links


class JPCMCIplus(PCMCI):
    r"""J-PCMCIplus causal discovery for time series datasets from multiple contexts.
        
    This class is based on the PCMCI framework as described in
    [i]. JPCMCIplus enables causal discovery for time series data from
    different contexts, i.e. datasets, where some of the variables
    describing the context might be unobserved. The method is described
    in detail in [ii]. See the tutorial for guidance in applying the
    method.

    References
    ----------
    .. [i] J. Runge, P. Nowack, M. Kretschmer, S. Flaxman, D. Sejdinovic,
       Detecting and quantifying causal associations in large nonlinear
       time series datasets. Sci. Adv. 5, eaau4996
       (2019) https://advances.sciencemag.org/content/5/11/eaau4996
    
    .. [ii] W. Günther, U. Ninad, J. Runge, Causal discovery for time
       series from multiple datasets with latent contexts. UAI 2023
    
    Parameters
    ----------
    node_classification : dictionary
        Classification of nodes into system, context, or dummy nodes.
        Keys of the dictionary are from {0, ..., N-1} where N is the number of nodes.
        Options for the values are "system", "time_context", "space_context", "time_dummy", or "space_dummy".
    
    Attributes
    ----------
    all_parents : dictionary
        Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
        the conditioning-parents estimated with PC algorithm.
    val_min : dictionary
        Dictionary of form val_min[j][(i, -tau)] = float
        containing the minimum test statistic value for each link estimated in
        the PC algorithm.
    pval_max : dictionary
        Dictionary of form pval_max[j][(i, -tau)] = float containing the maximum
        p-value for each link estimated in the PC algorithm.
    iterations : dictionary
        Dictionary containing further information on algorithm steps.
    N : int
        Number of variables.
    T : dict
        Time series sample length of dataset(s).
    dummy_parents : dictionary or None
        Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
        the dependence of the system nodes on the dummy nodes.
    observed_context_parents : dictionary or None
        Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
        the dependence of the system nodes on the observed context nodes.
    dummy_ci_test : conditional independence test object
        Conditional independence test used to test dependence between system nodes and dummy nodes.
        Currently, ParCorr is used with one-hot encoded dummies.
    mode : "system_search" or "context_search" or "dummy_search" (default: "system_search")
    time_context_nodes : list
        List with entries from {0, ..., N-1} where N is the number of nodes.
        This is the list of the temporal context nodes which are assumed to be constant over the different datasets.
    space_context_nodes :
        List with entries from {0, ..., N-1} where N is the number of nodes.
        This is the list of the spatial context nodes which are assumed to be constant over time.
    time_dummy : int or None (default: None)
        Node corresponding to the temporal dummy variable.
    space_dummy : int or None (default: None)
        Node corresponding to the spatial dummy variable.
    system_nodes : list
        List with entries from {0, ..., N-1} where N is the number of nodes.
        This is the list of the system nodes.
    """

    def __init__(self, node_classification, **kwargs):
        # Init base class
        PCMCI.__init__(self, **kwargs)

        self.system_nodes = self.group_nodes(node_classification, "system")
        self.time_context_nodes = self.group_nodes(node_classification, "time_context")
        self.space_context_nodes = self.group_nodes(node_classification, "space_context")
        self.time_dummy = self.group_nodes(node_classification, "time_dummy")
        self.space_dummy = self.group_nodes(node_classification, "space_dummy")

        self.dummy_parents = {i: [] for i in range(self.N)}
        self.observed_context_parents = {i: [] for i in range(self.N)}
        self.mode = "system_search"

    def group_nodes(self, node_types, node_type):
        nodes = range(self.N)
        return [node for node in nodes if node_types[node] == node_type]

    def run_jpcmciplus(self,
                       contemp_collider_rule='majority',
                       link_assumptions=None,
                       tau_min=0,
                       tau_max=1,
                       pc_alpha=0.01,
                       conflict_resolution=True,
                       reset_lagged_links=False,
                       max_conds_dim=None,
                       max_combinations=1,
                       max_conds_py=None,
                       max_conds_px=None,
                       max_conds_px_lagged=None,
                       fdr_method='none'):
        """Runs JPCMCIplus time-lagged and contemporaneous causal discovery for time series from multiple contexts.
        
        Method described in: W. Günther, U. Ninad, J. Runge, Causal discovery
        for time series from multiple datasets with latent contexts. UAI
        2023
        
        Notes
        -----
        The JPCMCIplus causal discovery method is described in [ii], where
        also analytical and numerical results are presented. JPCMCIplus can identify the joint causal graph
        over multiple datasets containing time series data from different contexts under the standard assumptions
        of Causal Sufficiency, Faithfulness and the Markov condition, as well as some background knowledge assumptions.
        JPCMCIplus estimates time-lagged and contemporaneous causal links from context to system
        variables and in between system variables by a four-step procedure:

        1.  **Discovery of supersets of the lagged parents of the system and observed temporal context nodes** by
        running the :math:`PC_1` lagged phase on this subset of nodes to obtain :math:`\\hat{\\mathcal{B}}^-_t(X_t^j)`.

        2.  Next, the **MCI test is run on pairs of system and context nodes conditional on subsets of system
        and context**, i.e. perform MCI tests for pairs :math:`((C^j_{t-\\tau}, X^i_t))_{\\tau > 0}`,
        :math:`(C_t^j, X_t^i)`, :math:`(X_t^i, C_t^j)` for all :math:`i,j`,

        .. math:: C_{t-\\tau}^i \\perp X_t^j | \\mathbf{S}, \\hat{\\mathcal{B}}^-_t(X_t^j)
                    \\setminus \\{ C_{t-\\tau}^i \\}, \\hat{\\mathcal{B}}^-_{t-\\tau}(C_{t-\\tau}^i)

        with :math:`\\mathbf{S}` being a subset of the contemporaneous adjacencies :math:`\\mathcal{A}_t(X_t^j)` and
        :math:`\\hat{\\mathcal{B}}^-_t(X_t^j)` are the lagged adjacencies from step one. If :math:`C` is a
        spatial context variable, we only have to test the contemporaneous pairs
        :math:`(C_t^j, X_t^i)`, :math:`(X_t^i, C_t^j)` for all :math:`i,j`.
        If :math:`C_t^j` and :math:`X_t^i` are conditionally independent, all lagged links between :math:`C_t^j` and
        :math:`X^j_{t-\\tau}` are also removed for all :math:`\\tau`.

        3.  **Perform MCI tests on all system-dummy pairs conditional on the superset of lagged links, the discovered
        contemporaneous context adjacencies, as well as on subsets of contemporaneous system links**, i.e. test
        for :math:`(D, X_t^i)`, :math:`(X_t^i, D)` for all :math:`i`, i.e.

        .. math:: D \\perp X_t^j | \\mathbf{S}, \\hat{\\mathcal{B}}^C_t(X_t^j),

        where :math:`\\mathbf{S} \\subset \\mathcal{A}_t(X_t^i)` and :math:`\\hat{\\mathcal{B}}^C_t(X_t^j)`
        are the lagged and contextual adjacencies found in the previous step.
        If :math:`D` and :math:`X_t^j` are found to be conditionally independence, links between :math:`D` and
        :math:`X^j_{t-\\tau}` are removed for all :math:`\\tau`.
        By assumption context node is the parent in all system-context links.

        4.  Finally, we **perform  MCI tests on all system pairs conditional on discovered lagged, context and dummy
        adjacencies, as well as on subsets of contemporaneous system links** and **orientation phase**. In more detail,
        we perform MCI test for pairs :math:`((X^j_{t-\\tau}, X_t^i))_{\\tau > 0}`, :math:`(X_t^i, X_t^j)` for all
        :math:`i, j`, i.e.

        .. math:: X^i_{t-\\tau} \\perp X_t^j | \\mathbf{S}, \\hat{\\mathcal{B}}^{CD}_t(X_t^j)
                    \\setminus \\{ X_{t-\\tau}^i \\},\\hat{\\mathcal{B}}^{CD}_t(X_{t-\\tau}^i)

        where :math:`\\mathbf{S} \\subset \\mathcal{A}_t(X_t^i)` and :math:`\\hat{\\mathcal{B}}^{CD}_t(X_t^j)`
        are the lagged, contextual, and dummy adjacencies found in the previous steps.
        Finally, all remaining edges (without expert knowledge) are oriented using the PCMCI+ orientation phase while
        making use of all triples involving one context or dummy variable and two system variables as in the non-time
        series case.

        JPCMCIplus can be flexibly combined with any kind of conditional
        independence test statistic adapted to the kind of data (continuous
        or discrete) and its assumed dependency types. These are available in
        ``tigramite.independence_tests``.
        See PCMCIplus for a description of the parameters of JPCMCIplus. Also, guidance on best practices for
        setting these parameters is given there.

        Parameters
        ----------
        link_assumptions : dict
            Dictionary of form {j:{(i, -tau): link_type, ...}, ...} specifying
            assumptions about links. This initializes the graph with entries
            graph[i,j,tau] = link_type. For example, graph[i,j,0] = '-->'
            implies that a directed link from i to j at lag 0 must exist.
            Valid link types are 'o-o', '-->', '<--'. In addition, the middle
            mark can be '?' instead of '-'. Then '-?>' implies that this link
            may not exist, but if it exists, its orientation is '-->'. Link
            assumptions need to be consistent, i.e., graph[i,j,0] = '-->'
            requires graph[j,i,0] = '<--' and acyclicity must hold. If a link
            does not appear in the dictionary, it is assumed absent. That is,
            if link_assumptions is not None, then all links have to be specified
            or the links are assumed absent.
        tau_min : int, optional (default: 0)
            Minimum time lag to test.
        tau_max : int, optional (default: 1)
            Maximum time lag. Must be larger or equal to tau_min.
        pc_alpha : float or list of floats, default: 0.01
            Significance level in algorithm. If a list or None is passed, the
            pc_alpha level is optimized for every graph across the given
            pc_alpha values ([0.001, 0.005, 0.01, 0.025, 0.05] for None) using
            the score computed in cond_ind_test.get_model_selection_criterion().
        contemp_collider_rule : {'majority', 'conservative', 'none'}
            Rule for collider phase to use. See the paper for details. Only
            'majority' and 'conservative' lead to an order-independent
            algorithm.
        conflict_resolution : bool, optional (default: True)
            Whether to mark conflicts in orientation rules. Only for True
            this leads to an order-independent algorithm.
        reset_lagged_links : bool, optional (default: False)
            Restricts the detection of lagged causal links in Step 2 to the
            significant adjacencies found in the PC1 algorithm in Step 1. For
            True, *all* lagged links are considered again, which improves
            detection power for lagged links, but also leads to larger
            runtimes.
        max_conds_dim : int, optional (default: None)
            Maximum number of conditions to test. If None is passed, this number
            is unrestricted.
        max_combinations : int, optional (default: 1)
            Maximum number of combinations of conditions of current cardinality
            to test. Defaults to 1 for PC_1 algorithm. For original PC algorithm
            a larger number, such as 10, can be used.
        max_conds_py : int, optional (default: None)
            Maximum number of lagged conditions of Y to use in MCI tests. If
            None is passed, this number is unrestricted.
        max_conds_px : int, optional (default: None)
            Maximum number of lagged conditions of X to use in MCI tests. If
            None is passed, this number is unrestricted.
        max_conds_px_lagged : int, optional (default: None)
            Maximum number of lagged conditions of X when X is lagged in MCI
            tests. If None is passed, this number is equal to max_conds_px.
        fdr_method : str, optional (default: 'none')
            Correction method, default is Benjamini-Hochberg False Discovery
            Rate method.
        Returns
        -------
        graph : array of shape [N, N, tau_max+1]
            Resulting causal graph, see description above for interpretation.
        val_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of test statistic values regarding adjacencies.
        p_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of p-values regarding adjacencies.
        sepset : dictionary
            Separating sets. See paper for details.
        ambiguous_triples : list
            List of ambiguous triples, only relevant for 'majority' and
            'conservative' rules, see paper for details.
        """
        observed_context_nodes = self.time_context_nodes + self.space_context_nodes

        # initialize / clean link_assumptions
        if link_assumptions is not None:
            _link_assumptions = deepcopy(link_assumptions)
        else:
            _link_assumptions = self._set_link_assumptions(link_assumptions, tau_min, tau_max,
                       remove_contemp=False)
            # {j: {(i, -tau): ("-?>" if tau > 0 else "o?o") for i in range(self.N)
            #                          for tau in range(tau_max + 1)} for j in range(self.N)}

        _link_assumptions = self.assume_exogenous_context(_link_assumptions, observed_context_nodes)
        _link_assumptions = self.clean_link_assumptions(_link_assumptions, tau_max)

        # for j in _link_assumptions:
        #     print(j, _link_assumptions[j])
        # self._set_link_assumptions(_link_assumptions, tau_min, tau_max,
        #                        remove_contemp=False)

        # Check if pc_alpha is chosen to optimize over a list
        if pc_alpha is None or isinstance(pc_alpha, (list, tuple, np.ndarray)):
            # Call optimizer wrapper around run_pcmciplus()
            return self._optimize_pcmciplus_alpha(
                link_assumptions=link_assumptions,
                tau_min=tau_min,
                tau_max=tau_max,
                pc_alpha=pc_alpha,
                contemp_collider_rule=contemp_collider_rule,
                conflict_resolution=conflict_resolution,
                reset_lagged_links=reset_lagged_links,
                max_conds_dim=max_conds_dim,
                max_combinations=max_combinations,
                max_conds_py=max_conds_py,
                max_conds_px=max_conds_px,
                max_conds_px_lagged=max_conds_px_lagged,
                fdr_method=fdr_method)

        elif pc_alpha < 0. or pc_alpha > 1:
            raise ValueError("Choose 0 <= pc_alpha <= 1")

        # Check the limits on tau
        self._check_tau_limits(tau_min, tau_max)

        # Step 0 and 1:
        context_results = self.discover_lagged_context_system_links(
            _link_assumptions,
            tau_min=tau_min,
            tau_max=tau_max,
            pc_alpha=pc_alpha,
            reset_lagged_links=reset_lagged_links,
            max_conds_dim=max_conds_dim,
            max_combinations=max_combinations,
            max_conds_py=max_conds_py,
            max_conds_px=max_conds_px,
            max_conds_px_lagged=max_conds_px_lagged,
            fdr_method=fdr_method
        )
        ctxt_res = deepcopy(context_results)
        # Store the parents in the pcmci member
        self.observed_context_parents = deepcopy(context_results['parents'])
        self.all_lagged_parents = deepcopy(
            context_results['lagged_parents'])  # remove context nodes from lagged parents
        self.all_lagged_parents = {i: [el for el in self.all_lagged_parents[i] if el[0] not in self.time_context_nodes]
                                   for i in
                                   range(self.N)}

        # if self.verbosity > 0:
        #     print("\nDiscovered observed context parents: ", context_results['parents'])

        if len(self.time_dummy) > 0 or len(self.space_dummy) > 0:
            # step 2:
            dummy_system_results = self.discover_dummy_system_links(
                _link_assumptions,
                ctxt_res,
                self.all_lagged_parents,
                tau_min=tau_min,
                tau_max=tau_max,
                pc_alpha=pc_alpha,
                reset_lagged_links=reset_lagged_links,
                max_conds_dim=max_conds_dim,
                max_conds_py=max_conds_py,
                max_conds_px=max_conds_px,
                max_conds_px_lagged=max_conds_px_lagged,
                fdr_method=fdr_method
            )
            # Store the parents in the pcmci member
            self.dummy_parents = dummy_system_results['parents']
        else:
            dummy_system_results = deepcopy(context_results)

        # if self.verbosity > 0:
        #     print("Discovered dummy parents: ", self.dummy_parents)

        # step 3:
        self.mode = "system_search"

        lagged_context_dummy_parents = {
            i: list(
                dict.fromkeys(self.all_lagged_parents[i] + self.observed_context_parents[i] + self.dummy_parents[i]))
            for i in self.system_nodes}
        # we only care about the parents of system nodes
        lagged_context_dummy_parents.update(
            {i: [] for i in observed_context_nodes + self.time_dummy + self.space_dummy})

        dummy_system_results_copy = deepcopy(dummy_system_results)

        # step 4:
        system_skeleton_results = self.discover_system_system_links(link_assumptions=_link_assumptions,
                                                                    lagged_context_dummy_parents=lagged_context_dummy_parents,
                                                                    tau_min=tau_min,
                                                                    tau_max=tau_max,
                                                                    pc_alpha=pc_alpha,
                                                                    reset_lagged_links=reset_lagged_links,
                                                                    max_conds_dim=max_conds_dim,
                                                                    max_conds_py=max_conds_py,
                                                                    max_conds_px=max_conds_px,
                                                                    max_conds_px_lagged=max_conds_px_lagged,
                                                                    fdr_method=fdr_method)

        # orientation phase
        colliders_step_results = self._pcmciplus_collider_phase(
            system_skeleton_results['graph'], system_skeleton_results['sepsets'],
            lagged_context_dummy_parents, pc_alpha,
            tau_min, tau_max, max_conds_py, max_conds_px, max_conds_px_lagged,
            conflict_resolution, contemp_collider_rule)

        final_graph = self._pcmciplus_rule_orientation_phase(colliders_step_results['graph'],
                                                             colliders_step_results['ambiguous_triples'],
                                                             conflict_resolution)

        # add context-system and dummy-system values and pvalues back in (lost because of link_assumption)
        for c in observed_context_nodes + self.time_dummy + self.space_dummy:
            for j in range(self.N):
                for lag in range(tau_max + 1):
                    # add context-system links to results
                    system_skeleton_results['val_matrix'][c, j, lag] = dummy_system_results_copy['val_matrix'][
                        c, j, lag]
                    system_skeleton_results['val_matrix'][j, c, lag] = dummy_system_results_copy['val_matrix'][
                        j, c, lag]

                    system_skeleton_results['p_matrix'][c, j, lag] = dummy_system_results_copy['p_matrix'][c, j, lag]
                    system_skeleton_results['p_matrix'][j, c, lag] = dummy_system_results_copy['p_matrix'][j, c, lag]

        # No confidence interval estimation here
        return_dict = {'graph': final_graph, 'p_matrix': system_skeleton_results['p_matrix'],
                       'val_matrix': system_skeleton_results['val_matrix'],
                       'sepsets': colliders_step_results['sepsets'],
                       'ambiguous_triples': colliders_step_results['ambiguous_triples'], 
                       'conf_matrix': None}

        # Print the results
        if self.verbosity > 0:
            self.print_results(return_dict, alpha_level=pc_alpha)

        # Return the dictionary
        self.results = return_dict

        # Return the dictionary
        return return_dict

    def assume_exogenous_context(self, link_assumptions, observed_context_nodes):
        """Helper function to amend the link_assumptions to ensure that all context-system links are oriented
        such that the context variable is the parent."""
        for j in link_assumptions:
            if j in self.system_nodes:
                for link in link_assumptions[j]:
                    i, lag = link
                    if i in observed_context_nodes + self.time_dummy + self.space_dummy:  # is context var
                        link_type = link_assumptions[j][link]
                        link_assumptions[j][link] = '-' + link_type[1] + '>'
        return link_assumptions

    def clean_link_assumptions(self, link_assumptions, tau_max):
        """Helper function to amend the link_assumptions in the following ways
            * remove any links where dummy is the child
            * remove any lagged links to dummy, and space_context (not to observed time context)
            * and system - context links where context is the child
            * and any links between spatial and temporal context
        """
        for node in self.time_dummy + self.space_dummy:
            link_assumptions[node] = {}

        for j in self.system_nodes + self.time_context_nodes + self.space_context_nodes:
            for lag in range(1, tau_max + 1):
                for c in self.time_dummy + self.space_dummy + self.space_context_nodes:
                    if (c, -lag) in link_assumptions[j]: link_assumptions[j].pop((c, -lag), None)
        for c in self.space_context_nodes + self.time_context_nodes:
            for j in self.system_nodes:
                for lag in range(tau_max + 1):
                    if (j, -lag) in link_assumptions[c]: link_assumptions[c].pop((j, -lag), None)
            if (c, 0) in link_assumptions[c]: link_assumptions[c].pop((c, 0), None)  # remove self-links

        for c in self.space_context_nodes:
            for k in self.time_context_nodes:
                for lag in range(tau_max + 1):
                    if (k, -lag) in link_assumptions[c]: link_assumptions[c].pop((k, -lag), None)
        for c in self.time_context_nodes:
            for k in self.space_context_nodes:
                if (k, 0) in link_assumptions[c]: link_assumptions[c].pop((k, 0), None)

        return link_assumptions

    def remove_dummy_link_assumptions(self, link_assumptions):
        """Helper function to remove any links to dummy from link_assumptions."""
        link_assumptions_wo_dummy = deepcopy(link_assumptions)
        for j in self.system_nodes + self.time_context_nodes + self.space_context_nodes:
            for dummy_node in self.time_dummy + self.space_dummy:
                if (dummy_node, 0) in link_assumptions_wo_dummy[j]:
                    link_assumptions_wo_dummy[j].pop((dummy_node, 0), None)
        return link_assumptions_wo_dummy

    def add_found_context_link_assumptions(self, link_assumptions, tau_max):
        """Helper function to add discovered links between system and observed context nodes to link_assumptions."""
        link_assumptions_dummy = deepcopy(link_assumptions)

        for c in self.space_context_nodes + self.time_context_nodes:
            link_assumptions_dummy[c] = {}
        for j in self.system_nodes + self.time_context_nodes + self.space_context_nodes:
            for c in self.space_context_nodes + self.time_context_nodes:
                for lag in range(tau_max + 1):
                    if (c, -lag) in link_assumptions_dummy[j]:
                        link_assumptions_dummy[j].pop((c, -lag), None)
            link_assumptions_dummy[j].update({parent: '-->' for parent in self.observed_context_parents[j]})

        return link_assumptions_dummy

    def clean_system_link_assumptions(self, link_assumptions, tau_max):
        """Helper function to remove any links to dummy and observed context nodes from link_assumptions.
            Add discovered links to contextual parents (from steps 1 and 2) to the link_assumptions.
        """
        dummy_vars = self.time_dummy + self.space_dummy
        observed_context_nodes = self.time_context_nodes + self.space_context_nodes
        system_links = deepcopy(link_assumptions)

        for j in self.system_nodes:
            for C in dummy_vars + observed_context_nodes:
                for lag in range(tau_max + 1):
                    if (C, -lag) in system_links[j]:
                        system_links[j].pop((C, -lag), None)

        for j in system_links:
            system_links[j].update(
                {parent: '-->' for parent in self.observed_context_parents[j] + self.dummy_parents[j]})

        for C in observed_context_nodes + dummy_vars:
            # we are not interested in links between context variables (thus system_links[C] = {})
            system_links[C] = {}
        return system_links

    def discover_lagged_context_system_links(self, link_assumptions,
                                             tau_min=0,
                                             tau_max=1, pc_alpha=0.01,
                                             reset_lagged_links=False,
                                             max_conds_dim=None,
                                             max_combinations=1,
                                             max_conds_py=None,
                                             max_conds_px=None,
                                             max_conds_px_lagged=None,
                                             fdr_method='none'):
        """
        Step 1 of JPCMCIplus, i.e. discovery of links between observed context nodes and system nodes through an
        application of the skeleton phase of PCMCIplus to this subset of nodes (observed context nodes and system
        nodes).
        See run_jpcmciplus for a description of the parameters.

        Returns
        -------
        graph : array of shape [N, N, tau_max+1]
            Resulting causal graph, see description above for interpretation.
        val_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of test statistic values regarding adjacencies.
        p_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of p-values regarding adjacencies.
        parents : dictionary
            Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
            the estimated context parents of the system nodes.
        lagged_parents : dictionary
            Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
            the conditioning-parents estimated with PC algorithm.
        """

        # Initializing
        context_parents = {i: [] for i in range(self.N)}

        # find links between expressive context, and between expressive context and system nodes
        # here, we exclude any links to dummy
        _link_assumptions_wo_dummy = self.remove_dummy_link_assumptions(link_assumptions)
        _int_link_assumptions = self._set_link_assumptions(_link_assumptions_wo_dummy, tau_min, tau_max)

        # Step 1: Get a superset of lagged parents from run_pc_stable
        if self.verbosity > 0:
            print("\n##\n## J-PCMCI+ Step 1: Selecting lagged conditioning sets\n##")

        lagged_parents = self.run_pc_stable(link_assumptions=link_assumptions,
                                            tau_min=tau_min,
                                            tau_max=tau_max,
                                            pc_alpha=pc_alpha,
                                            max_conds_dim=max_conds_dim,
                                            max_combinations=max_combinations)

        self.mode = "context_search"

        p_matrix = self.p_matrix
        val_matrix = self.val_matrix

        # run PCMCI+ skeleton phase on subset of links to discover context-system links
        if self.verbosity > 0:
            print("\n##\n## J-PCMCI+ Step 2: Discovering context-system links\n##")
            if link_assumptions is not None:
                print("\nWith link_assumptions = %s" % str(_int_link_assumptions))

        skeleton_results = self._pcmciplus_mci_skeleton_phase(
            lagged_parents, _int_link_assumptions, pc_alpha,
            tau_min, tau_max, max_conds_dim, None,
            max_conds_py, max_conds_px, max_conds_px_lagged,
            reset_lagged_links, fdr_method,
            p_matrix, val_matrix
        )

        skeleton_graph = skeleton_results['graph']

        for j in self.system_nodes + self.time_context_nodes + self.space_context_nodes:
            for c in self.space_context_nodes + self.time_context_nodes:
                for k in range(tau_max + 1):
                    if skeleton_graph[c, j, k] == 'o?o' or skeleton_graph[c, j, k] == '-?>' or skeleton_graph[
                        c, j, k] == 'o-o' or skeleton_graph[c, j, k] == '-->':
                        context_parents[j].append((c, -k))

        return_dict = {'graph': skeleton_results['graph'], 'p_matrix': skeleton_results['p_matrix'],
                       'val_matrix': skeleton_results['val_matrix'],
                       'parents': context_parents, 'lagged_parents': lagged_parents}

        # Print the results
        if self.verbosity > 0:
            self.print_results(return_dict, alpha_level=pc_alpha)

        return return_dict

    def discover_dummy_system_links(self, link_assumptions,
                                    context_system_results,
                                    lagged_parents,
                                    tau_min=0,
                                    tau_max=1,
                                    pc_alpha=0.01,
                                    reset_lagged_links=False,
                                    max_conds_dim=None,
                                    max_conds_py=None,
                                    max_conds_px=None,
                                    max_conds_px_lagged=None,
                                    fdr_method='none'):
        """
        Step 2 of JPCMCIplus, i.e. discovery of links between observed (time and space) dummy nodes and system nodes
        through an application of the skeleton phase of PCMCIplus to this subset of nodes (dummy nodes and
        system nodes).
        See run_jpcmciplus for a description of the parameters.

        Parameters
        ----------
        context_system_results : dictionary
            Output of discover_lagged_and_context_system_links, i.e. lagged and context parents together with the
            corresponding estimated test statistic values regarding adjacencies.
        lagged_parents : dictionary
            Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing the conditioning-parents
            estimated with PC algorithm.

        Returns
        -------
        graph : array of shape [N, N, tau_max+1]
            Resulting causal graph, see description above for interpretation.
        val_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of test statistic values regarding adjacencies.
        p_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of p-values regarding adjacencies.
        parents : dictionary
            Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
            the estimated dummy parents of the system nodes.
        """
        lagged_context_parents = {i: list(dict.fromkeys(context_system_results['parents'][i] + lagged_parents[i])) for
                                  i in range(self.N)}
        dummy_parents = {i: [] for i in range(self.N)}
        p_matrix = context_system_results['p_matrix']

        # setup link assumptions without the observed context nodes
        _link_assumptions_dummy = self.add_found_context_link_assumptions(link_assumptions, tau_max)
        _int_link_assumptions = self._set_link_assumptions(_link_assumptions_dummy, tau_min, tau_max)

        self.mode = "dummy_search"
        if self.verbosity > 0:
            print("\n##\n## J-PCMCI+ Step 3: Discovering dummy-system links\n##")
            if _link_assumptions_dummy is not None:
                print("\nWith link_assumptions = %s" % str(_int_link_assumptions))

        skeleton_results_dummy = self._pcmciplus_mci_skeleton_phase(
            lagged_context_parents, _int_link_assumptions, pc_alpha,
            tau_min, tau_max, max_conds_dim, None,
            max_conds_py, max_conds_px, max_conds_px_lagged,
            reset_lagged_links, fdr_method,
            self.p_matrix, self.val_matrix
        )

        skeleton_graph_dummy = skeleton_results_dummy['graph']

        for j in self.system_nodes:
            for k in range(tau_max + 1):
                for dummy_node in self.time_dummy + self.space_dummy:
                    if skeleton_graph_dummy[dummy_node, j, k] == 'o?o' or \
                            skeleton_graph_dummy[dummy_node, j, k] == '-?>' or \
                            skeleton_graph_dummy[dummy_node, j, k] == 'o-o' or \
                            skeleton_graph_dummy[dummy_node, j, k] == '-->':
                        dummy_parents[j].append((dummy_node, k))
                for context_node in self.time_context_nodes + self.space_context_nodes:
                    skeleton_results_dummy['val_matrix'][context_node, j, k] = context_system_results['val_matrix'][
                        context_node, j, k]
                    skeleton_results_dummy['val_matrix'][j, context_node, k] = context_system_results['val_matrix'][
                        j, context_node, k]

                    skeleton_results_dummy['p_matrix'][context_node, j, k] = p_matrix[context_node, j, k]
                    skeleton_results_dummy['p_matrix'][j, context_node, k] = p_matrix[j, context_node, k]

        return_dict = {'graph': skeleton_results_dummy['graph'], 'p_matrix': skeleton_results_dummy['p_matrix'],
                       'val_matrix': skeleton_results_dummy['val_matrix'], 'parents': dummy_parents}

        # Print the results
        if self.verbosity > 0:
            self.print_results(return_dict, alpha_level=pc_alpha)
        return return_dict

    def discover_system_system_links(self, link_assumptions,
                                     lagged_context_dummy_parents,
                                     tau_min=0,
                                     tau_max=1,
                                     pc_alpha=0.01,
                                     reset_lagged_links=False,
                                     max_conds_dim=None,
                                     max_conds_py=None,
                                     max_conds_px=None,
                                     max_conds_px_lagged=None,
                                     fdr_method='none'
                                     ):
        """
        Step 4 of JPCMCIplus and orientation phase, i.e. discovery of links between system nodes given the knowledge
        about their context parents through an application of PCMCIplus to this subset of nodes (system nodes).
        See run_jpcmciplus for a description of the other parameters.


        Parameters
        ----------
        lagged_context_dummy_parents : dictionary
            Dictionary containing lagged and (dummy and observed) context parents of the system nodes estimated during
            step 1 and step 2 of J-PCMCI+.

        Returns
        -------
        graph : array of shape [N, N, tau_max+1]
            Resulting causal graph, see description above for interpretation.
        val_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of test statistic values regarding adjacencies.
        p_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of p-values regarding adjacencies.
        sepset : dictionary
            Separating sets. See paper for details.
        ambiguous_triples : list
            List of ambiguous triples, only relevant for 'majority' and
            'conservative' rules, see paper for details.
        """
        self.mode = "system_search"

        # Get the parents from run_pc_stable only on the system links
        system_links = self.clean_system_link_assumptions(link_assumptions, tau_max)
        # Set the selected links
        _int_link_assumptions = self._set_link_assumptions(system_links, tau_min, tau_max)

        if self.verbosity > 0:
            print("\n##\n## J-PCMCI+ Step 4: Discovering system-system links \n##")
            if system_links is not None:
                print("\nWith link_assumptions = %s" % str(_int_link_assumptions))

        skeleton_results = self._pcmciplus_mci_skeleton_phase(
            lagged_context_dummy_parents, _int_link_assumptions, pc_alpha,
            tau_min, tau_max, max_conds_dim, None,
            max_conds_py, max_conds_px, max_conds_px_lagged,
            reset_lagged_links, fdr_method,
            self.p_matrix, self.val_matrix
        )

        return skeleton_results

    def _remaining_pairs(self, graph, adjt, tau_min, tau_max, p):
        """Helper function returning the remaining pairs that still need to be
        tested depending on the JPCMCIplus step, i.e. discovery of context-system links (step 1),
        dummy-context links (step 2) or system-system links in which case the function of the parent class is called.
        """
        all_context_nodes = self.time_context_nodes + self.space_context_nodes
        if self.mode == "context_search":
            # during discovery of context-system links we are only
            # interested in context-context and context-system pairs
            N = graph.shape[0]
            pairs = []
            for (i, j) in itertools.product(range(N), range(N)):
                for abstau in range(tau_min, tau_max + 1):
                    if (graph[i, j, abstau] != ""
                            and len(
                                [a for a in adjt[j] if a != (i, -abstau)]) >= p
                            and i in all_context_nodes):
                        pairs.append((i, j, abstau))
            return pairs
        elif self.mode == "dummy_search":
            # during discovery of dummy-system links we are only
            # interested in dummy-system pairs
            N = graph.shape[0]
            pairs = []
            for (i, j) in itertools.product(range(N), range(N)):
                for abstau in range(tau_min, tau_max + 1):
                    if (graph[i, j, abstau] != ""
                            and len(
                                [a for a in adjt[j] if a != (i, -abstau)]) >= p
                            and i in self.time_dummy + self.space_dummy):
                        pairs.append((i, j, abstau))
            return pairs
        else:
            return super()._remaining_pairs(graph, adjt, tau_min, tau_max, p)

    def _run_pcalg_test(self, graph, i, abstau, j, S, lagged_parents, max_conds_py,
                        max_conds_px, max_conds_px_lagged, tau_max, alpha_or_thres=None):
        """MCI conditional independence tests within PCMCIplus or PC algorithm. Depending on the JPCMCIplus step
        the setup is adapted slightly. During the discovery of dummy-system links (step 2) we are using
        the dummy_ci_test and condition on the parents found during step 1; during the discovery of system-system links
        (step 3) we are conditioning on the found contextual parents.
        Parameters
        ----------
        graph : array
            ...
        i : int
            Variable index.
        abstau : int
            Time lag (absolute value).
        j : int
            Variable index.
        S : list
            List of contemporaneous conditions.
        lagged_parents : dictionary of lists
            Dictionary of lagged parents for each node.
        max_conds_py : int
            Max number of lagged parents for node j.
        max_conds_px : int
            Max number of lagged parents for lagged node i.
        max_conds_px_lagged : int
            Maximum number of lagged conditions of X when X is lagged in MCI
            tests. If None is passed, this number is equal to max_conds_px.
        tau_max : int
            Maximum time lag.
        Returns
        -------
        val : float
            Test statistic value.
        pval : float
            Test statistic p-value.
        Z : list
            List of conditions.
        """

        if self.mode == 'dummy_search':
            # during discovery of dummy-system links we condition on the found contextual parents from step 1.
            cond = list(S) + self.observed_context_parents[j]
            cond = list(dict.fromkeys(cond))  # remove overlapps
            return super()._run_pcalg_test(graph, i, abstau, j, cond, lagged_parents, max_conds_py,
                                           max_conds_px, max_conds_px_lagged, tau_max, alpha_or_thres)

        elif self.mode == 'system_search':
            # during discovery of system-system links we are conditioning on the found contextual parents
            cond = list(S) + self.dummy_parents[j] + self.observed_context_parents[j]
            cond = list(dict.fromkeys(cond))  # remove overlapps
            return super()._run_pcalg_test(graph, i, abstau, j, cond, lagged_parents, max_conds_py,
                                           max_conds_px, max_conds_px_lagged, tau_max, alpha_or_thres)
        else:
            return super()._run_pcalg_test(graph, i, abstau, j, S, lagged_parents, max_conds_py,
                                           max_conds_px, max_conds_px_lagged, tau_max, alpha_or_thres)

if __name__ == '__main__':
    # Imports
    from numpy.random import SeedSequence, default_rng

    import tigramite
    from tigramite.toymodels import structural_causal_processes as toys
    from tigramite.toymodels.context_model import ContextModel
    # from tigramite.jpcmciplus import JPCMCIplus
    from tigramite.independence_tests.parcorr_mult import ParCorrMult
    import tigramite.data_processing as pp
    import tigramite.plotting as tp


    # Set seeds for reproducibility
    ss = SeedSequence(12345)
    child_seeds = ss.spawn(2)

    model_seed = child_seeds[0]
    context_seed = child_seeds[1]

    random_state = np.random.default_rng(model_seed)

    # Choose the time series length and number of spatial contexts
    T = 100
    nb_domains = 50

    transient_fraction=0.2
    tau_max = 2
    frac_observed = 0.5

    # Specify the model
    def lin(x): return x

    links = {0: [((0, -1), 0.3, lin), ((3, -1), 0.7, lin), ((4, 0), 0.9, lin)],
             1: [((1, -1), 0.4, lin), ((3, -1), 0.8, lin)],
             2: [((2, -1), 0.3, lin), ((1, 0), -0.5, lin), ((4, 0), 0.5, lin), ((5, 0), 0.6, lin)] ,
             3: [], 
             4: [], 
             5: []
                }

    # Specify which node is a context node via node_type (can be "system", "time_context", or "space_context")
    node_classification = {
        0: "system",
        1: "system",
        2: "system",
        3: "time_context",
        4: "time_context",
        5: "space_context"
    }

    # Specify dynamical noise term distributions, here unit variance Gaussians
    #random_state = np.random.RandomState(seed)
    noises = [random_state.standard_normal for j in range(6)]

    contextmodel = ContextModel(links=links, node_classification=node_classification,
                                noises=noises, 
                                seed=context_seed)

    data_ens, nonstationary = contextmodel.generate_data(nb_domains, T)

    assert not nonstationary

    system_indices = [0,1,2]
    # decide which context variables should be latent, and which are observed
    observed_indices_time = [4]
    latent_indices_time = [3]

    observed_indices_space = [5]
    latent_indices_space = []

    # all system variables are also observed, thus we get the following observed data
    observed_indices = system_indices + observed_indices_time + observed_indices_space
    data_observed = {key: data_ens[key][:,observed_indices] for key in data_ens}


    # Add one-hot-encoding of time-steps and dataset index to the observational data. 
    # These are the values of the time and space dummy variables.
    dummy_data_time = np.identity(T)

    data_dict = {}
    for i in range(nb_domains):
        dummy_data_space = np.zeros((T, nb_domains))
        dummy_data_space[:, i] = 1.
        data_dict[i] = np.hstack((data_observed[i], dummy_data_time, dummy_data_space))

    # Define vector-valued variables including dummy variables as well as observed (system and context) variables
    nb_observed_context_nodes = len(observed_indices_time) + len(observed_indices_space)
    N = len(system_indices)
    process_vars = system_indices
    observed_temporal_context_nodes = list(range(N, N + len(observed_indices_time)))
    observed_spatial_context_nodes = list(range(N + len(observed_indices_time), 
                                                N + len(observed_indices_time) + len(observed_indices_space)))
    time_dummy_index = N + nb_observed_context_nodes
    space_dummy_index = N + nb_observed_context_nodes + 1
    time_dummy = list(range(time_dummy_index, time_dummy_index + T))
    space_dummy = list(range(time_dummy_index + T, time_dummy_index + T + nb_domains))

    vector_vars = {i: [(i, 0)] for i in process_vars + observed_temporal_context_nodes + observed_spatial_context_nodes}
    vector_vars[time_dummy_index] = [(i, 0) for i in time_dummy]
    vector_vars[space_dummy_index] = [(i, 0) for i in space_dummy]

    # Name all the variables and initialize the dataframe object
    # Be careful to use analysis_mode = 'multiple'
    sys_var_names = ['X_' + str(i) for i in process_vars]
    context_var_names = ['t-C_'+str(i) for i in observed_indices_time] + ['s-C_'+str(i) for i in observed_indices_space]
    var_names = sys_var_names + context_var_names + ['t-dummy', 's-dummy']

    dataframe = pp.DataFrame(
        data=data_dict,
        vector_vars = vector_vars,
        analysis_mode = 'multiple',
        var_names = var_names
        )


    # Classify all the nodes into system, context, or dummy
    node_classification_jpcmci = {i: node_classification[var] for i, var in enumerate(observed_indices)}
    node_classification_jpcmci.update({time_dummy_index : "time_dummy", space_dummy_index : "space_dummy"})

    # Create a J-PCMCI+ object, passing the dataframe and (conditional)
    # independence test objects, as well as the observed temporal and spatial context nodes 
    # and the indices of the dummies.
    JPCMCIplus = JPCMCIplus(dataframe=dataframe,
                              cond_ind_test=ParCorrMult(significance='analytic'), 
                              node_classification=node_classification_jpcmci,
                              verbosity=1,)

    # Define the analysis parameters.
    tau_max = 2
    pc_alpha = 0.01

    # Run J-PCMCI+
    results = JPCMCIplus.run_jpcmciplus(tau_min=0, 
                                  tau_max=tau_max, 
                                  pc_alpha=pc_alpha)


"""Tigramite causal discovery for time series."""

# Author: Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

from __future__ import print_function
import numpy as np

from collections import defaultdict, OrderedDict
from itertools import combinations, permutations


class OracleCI:
    r"""Oracle of conditional independence test X _|_ Y | Z given a graph.

    Class around link_coeff causal ground truth. X _|_ Y | Z is based on
    assessing whether X and Y are d-separated given Z in the graph.

    Class can be used just like a Tigramite conditional independence class
    (e.g., ParCorr). The main use is for unit testing of PCMCI methods.

    Parameters
    ----------
    graph : array of shape [N, N, tau_max+1]
        Causal graph.
    links : dict
        Dictionary of form {0:[(0, -1), ...], 1:[...], ...}.
        Alternatively can also digest {0: [((0, -1), coeff, func)], ...}.
    observed_vars : None or list, optional (default: None)
        Subset of keys in links definining which variables are 
        observed. If None, then all variables are observed.
    selection_vars : None or list, optional (default: None)
        Subset of keys in links definining which variables are 
        selected (= always conditioned on at every time lag).
        If None, then no variables are selected.
    verbosity : int, optional (default: 0)
        Level of verbosity.
    """

    # documentation
    @property
    def measure(self):
        """
        Concrete property to return the measure of the independence test
        """
        return self._measure

    def __init__(self,
                 links=None,
                 observed_vars=None,
                 selection_vars=None,
                 graph=None,
                 graph_is_mag=False,
                 tau_max=None,
                 verbosity=0):

        self.tau_max = tau_max
        self.graph_is_mag = graph_is_mag

        if links is None:
            if graph is None:
                raise ValueError("Either links or graph must be specified!")
            else:
                # Get canonical DAG from graph, potentially interpreted as MAG
                # self.tau_max = graph.shape[2]
                (links, 
                 observed_vars, 
                 selection_vars) = self.get_links_from_graph(graph)
                # # TODO make checks and tau_max?
                # self.graph = graph


        self.verbosity = verbosity
        self._measure = 'oracle_ci'
        self.confidence = None
        self.links = links
        self.N = len(links)
        # self.tau_max = self._get_minmax_lag(self.links)

        # Initialize already computed dsepsets of X, Y, Z
        self.dsepsets = {}

        # Initialize observed vars
        self.observed_vars = observed_vars
        if self.observed_vars is None:
            self.observed_vars = range(self.N)
        else:
            if not set(self.observed_vars).issubset(set(range(self.N))):
                raise ValueError("observed_vars must be subset of range(N).")
            if self.observed_vars != sorted(self.observed_vars):
                raise ValueError("observed_vars must ordered.")
            if len(self.observed_vars) != len(set(self.observed_vars)):
                raise ValueError("observed_vars must not contain duplicates.")

        self.selection_vars = selection_vars

        if self.selection_vars is not None:
            if not set(self.selection_vars).issubset(set(range(self.N))):
                raise ValueError("selection_vars must be subset of range(N).")
            if self.selection_vars != sorted(self.selection_vars):
                raise ValueError("selection_vars must ordered.")
            if len(self.selection_vars) != len(set(self.selection_vars)):
                raise ValueError("selection_vars must not contain duplicates.")
        else:
            self.selection_vars = []

        # ToDO: maybe allow to use user-tau_max, otherwise deduced from links
        self.graph = self.get_graph_from_links(tau_max=tau_max)

        self.ci_results = {}

    def set_dataframe(self, dataframe):
        """Dummy function."""
        pass

    def _check_XYZ(self, X, Y, Z):
        """Checks variables X, Y, Z.

        Parameters
        ----------
        X, Y, Z : list of tuples
            For a dependence measure I(X;Y|Z), Y is of the form [(varY, 0)],
            where var specifies the variable index. X typically is of the form
            [(varX, -tau)] with tau denoting the time lag and Z can be
            multivariate [(var1, -lag), (var2, -lag), ...] .

        Returns
        -------
        X, Y, Z : tuple
            Cleaned X, Y, Z.
        """
        # Get the length in time and the number of nodes
        N = self.N

        # Remove duplicates in X, Y, Z
        X = list(OrderedDict.fromkeys(X))
        Y = list(OrderedDict.fromkeys(Y))
        Z = list(OrderedDict.fromkeys(Z))

        # If a node in Z occurs already in X or Y, remove it from Z
        Z = [node for node in Z if (node not in X) and (node not in Y)]

        # Check that all lags are non-positive and indices are in [0,N-1]
        XYZ = X + Y + Z
        dim = len(XYZ)
        # Ensure that XYZ makes sense
        if np.array(XYZ).shape != (dim, 2):
            raise ValueError("X, Y, Z must be lists of tuples in format"
                             " [(var, -lag),...], eg., [(2, -2), (1, 0), ...]")
        if np.any(np.array(XYZ)[:, 1] > 0):
            raise ValueError("nodes are %s, " % str(XYZ) +
                             "but all lags must be non-positive")
        if (np.any(np.array(XYZ)[:, 0] >= N)
                or np.any(np.array(XYZ)[:, 0] < 0)):
            raise ValueError("var indices %s," % str(np.array(XYZ)[:, 0]) +
                             " but must be in [0, %d]" % (N - 1))
        if np.all(np.array(Y)[:, 1] != 0):
            raise ValueError("Y-nodes are %s, " % str(Y) +
                             "but one of the Y-nodes must have zero lag")

        return (X, Y, Z)

    def _get_lagged_parents(self, var_lag, exclude_contemp=False,
                only_non_causal_paths=False, X=None, causal_children=None):
        """Helper function to yield lagged parents for var_lag from
        self.links_coeffs.

        Parameters
        ----------
        var_lag : tuple
            Tuple of variable and lag which is assumed <= 0.
        exclude_contemp : bool
            Whether contemporaneous links should be exluded.

        Yields
        ------
        Next lagged parent.
        """

        var, lag = var_lag

        for link_props in self.links[var]:
            if len(link_props) == 3:
                i, tau = link_props[0]
                coeff = link_props[1]
            else:
                i, tau = link_props
                coeff = 1.
            if coeff != 0.:
                if not (exclude_contemp and lag == 0):
                    if only_non_causal_paths:
                        if not ((i, lag + tau) in X and var_lag in causal_children):
                            yield (i, lag + tau)
                    else:
                        yield (i, lag + tau)

    def _get_children(self):
        """Helper function to get children from links.

        Note that for children the lag is positive.

        Returns
        -------
        children : dict
            Dictionary of form {0:[(0, 1), (3, 0), ...], 1:[], ...}.
        """

        N = len(self.links)
        children = dict([(j, []) for j in range(N)])

        for j in range(N):
            for link_props in self.links[j]:
                if len(link_props) == 3:
                    i, tau = link_props[0]
                    coeff = link_props[1]
                else:
                    i, tau = link_props
                    coeff = 1.
                if coeff != 0.:
                    children[i].append((j, abs(tau)))

        return children

    def _get_lagged_children(self, var_lag, children, exclude_contemp=False,
           only_non_causal_paths=False, X=None, causal_children=None):
        """Helper function to yield lagged children for var_lag from children.

        Parameters
        ----------
        var_lag : tuple
            Tuple of variable and lag which is assumed <= 0.
        children : dict
            Dictionary of form {0:[(0, 1), (3, 0), ...], 1:[], ...}.
        exclude_contemp : bool
            Whether contemporaneous links should be exluded.

        Yields
        ------
        Next lagged child.
        """

        var, lag = var_lag
        # lagged_parents = []

        for child in children[var]:
            k, tau = child
            if not (exclude_contemp and tau == 0):
                # lagged_parents.append((i, lag + tau))
                if only_non_causal_paths:
                    if not (var_lag in X and (k, lag + tau) in causal_children):
                        yield (k, lag + tau)
                else:
                    yield (k, lag + tau)

    def _get_non_blocked_ancestors(self, Y, conds=None, mode='non_repeating',
                                    max_lag=None):
        """Helper function to return the non-blocked ancestors of variables Y.

        Returns a dictionary of ancestors for every y in Y. y is a tuple (
        var, lag) where lag <= 0. All ancestors with directed paths towards y
        that are not blocked by conditions in conds are included. In mode
        'non_repeating' an ancestor X^i_{t-\tau_i} with link X^i_{t-\tau_i}
        --> X^j_{ t-\tau_j} is only included if X^i_{t'-\tau_i} --> X^j_{
        t'-\tau_j} is not already part of the ancestors. The most lagged
        ancestor for every variable X^i defines the maximum ancestral time
        lag, which is also returned. In mode 'max_lag' ancestors are included
        up to the maximum time lag max_lag.

        It's main use is to return the maximum ancestral time lag max_lag of
        y in Y for every variable in self.links_coeffs.

        Parameters
        ----------
        Y : list of tuples
            Of the form [(var, -tau)], where var specifies the variable
            index and tau the time lag.
        conds : list of tuples
            Of the form [(var, -tau)], where var specifies the variable
            index and tau the time lag.
        mode : {'non_repeating', 'max_lag'}
            Whether repeating links should be excluded or ancestors should be
            followed up to max_lag.
        max_lag : int
            Maximum time lag to include ancestors.

        Returns
        -------
        ancestors : dict
            Includes ancestors for every y in Y.
        max_lag : int
            Maximum time lag to include ancestors.
        """

        def _repeating(link, seen_links):
            """Returns True if a link or its time-shifted version is already
            included in seen_links."""
            i, taui = link[0]
            j, tauj = link[1]

            for seen_link in seen_links:
                seen_i, seen_taui = seen_link[0]
                seen_j, seen_tauj = seen_link[1]

                if (i == seen_i and j == seen_j
                    and abs(tauj-taui) == abs(seen_tauj-seen_taui)):
                    return True

            return False

        if conds is None:
            conds = []

        conds = [z for z in conds if z not in Y]

        N = len(self.links)

        # Initialize max. ancestral time lag for every N
        if mode == 'non_repeating':
            max_lag = 0
        else:
            if max_lag is None:
                raise ValueError("max_lag must be set in mode = 'max_lag'")

        if self.selection_vars is not None:
            for selection_var in self.selection_vars:
                # print (selection_var, conds)
                # print([(selection_var, -tau_sel) for tau_sel in range(0, max_lag + 1)])
                conds += [(selection_var, -tau_sel) for tau_sel in range(0, max_lag + 1)]

        ancestors = dict([(y, []) for y in Y])

        for y in Y:
            j, tau = y   # tau <= 0
            if mode == 'non_repeating':
                max_lag = max(max_lag, abs(tau))
            seen_links = []
            this_level = [y]
            while len(this_level) > 0:
                next_level = []
                for varlag in this_level:
                    for par in self._get_lagged_parents(varlag):
                        i, tau = par
                        if par not in conds and par not in ancestors[y]:
                            if ((mode == 'non_repeating' and
                                not _repeating((par, varlag), seen_links)) or
                                (mode == 'max_lag' and
                                 abs(tau) <= abs(max_lag))):
                                    ancestors[y].append(par)
                                    if mode == 'non_repeating':
                                        max_lag = max(max_lag,
                                                         abs(tau))
                                    next_level.append(par)
                                    seen_links.append((par, varlag))

                this_level = next_level

        return ancestors, max_lag

    def _get_maximum_possible_lag(self, XYZ):
        """Helper function to return the maximum time lag of any confounding path.

        This is still based on a conjecture!

        The conjecture states that if and only if X and Y are d-connected given Z
        in a stationary DAG, then there exists a confounding path with a maximal
        time lag (i.e., the node on that path with maximal lag) given as follows:
        For any node in XYZ consider all non-repeating causal paths from the past
        to that node, where non-repeating means that a link X^i_{t-\tau_i}
        --> X^j_{ t-\tau_j} is only traversed if X^i_{t'-\tau_i} --> X^j_{
        t'-\tau_j} is not already part of that path. The most lagged
        ancestor for every variable node in XYZ defines the maximum ancestral time
        lag, which is returned.

        Parameters
        ----------
        XYZ : list of tuples
            Of the form [(var, -tau)], where var specifies the variable
            index and tau the time lag.

        Returns
        -------
        max_lag : int
            Maximum time lag of non-repeating causal path ancestors.
        """

        def _repeating(link, seen_path):
            """Returns True if a link or its time-shifted version is already
            included in seen_links."""
            i, taui = link[0]
            j, tauj = link[1]

            for index, seen_link in enumerate(seen_path[:-1]):
                seen_i, seen_taui = seen_link
                seen_j, seen_tauj = seen_path[index + 1]

                if (i == seen_i and j == seen_j
                    and abs(tauj-taui) == abs(seen_tauj-seen_taui)):
                    return True

            return False

        N = len(self.links)

        # Initialize max. ancestral time lag for every N
        max_lag = 0
    
        # Not sure whether this is relevant!
        # if self.selection_vars is not None:
        #     for selection_var in self.selection_vars:
        #         # print (selection_var, conds)
        #         # print([(selection_var, -tau_sel) for tau_sel in range(0, max_lag + 1)])
        #         conds += [(selection_var, -tau_sel) for tau_sel in range(0, max_lag + 1)]

        # ancestors = dict([(y, []) for y in Y])

        for y in XYZ:
            j, tau = y   # tau <= 0
            max_lag = max(max_lag, abs(tau))
                
            causal_path = []
            queue = [(y, causal_path)]

            while queue:
                varlag, causal_path = queue.pop()
                causal_path = [varlag] + causal_path

                for node in self._get_lagged_parents(varlag):
                    i, tau = node

                    if (node not in causal_path):
                    
                        if len(causal_path) == 1:
                            queue.append((node, causal_path))
                            continue

                        if (len(causal_path) > 1) and not _repeating((node, varlag), causal_path):
                            
                                max_lag = max(max_lag, abs(tau))
                                queue.append((node, causal_path))

        if self.verbosity > 0:
            print("Max. non-repeated ancestral time lag: ", max_lag)

        # ATTENTION: this may not find correct common ancestors, therefore multiply by 10
        # until the problem is solved
        max_lag *= 10

        return max_lag

    def _get_descendants(self, W, children, max_lag, ignore_time_bounds=False):
        """Get descendants of nodes in W up to time t.
        
        Includes the nodes themselves.
        """

        descendants = set(W)

        for w in W:
            j, tau = w 
            this_level = [w]
            while len(this_level) > 0:
                next_level = []
                for varlag in this_level:
                    for child in self._get_lagged_children(varlag, children):
                        i, tau = child
                        if (child not in descendants 
                            and (-max_lag <= tau <= 0 or ignore_time_bounds)):
                            descendants = descendants.union(set([child]))
                            next_level.append(child)

                this_level = next_level       

        return list(descendants)

    def _has_any_path(self, X, Y, conds, max_lag=None, 
        starts_with=None, ends_with=None,
        directed=False,
        forbidden_nodes=None,
        only_non_causal_paths=False,
        check_optimality_cond=False,
        optimality_cond_des_YM=None,
        optimality_cond_Y=None,
        only_collider_paths_with_vancs=False,
        XYS=None,
        return_path=False):
        """Returns True if X and Y are d-connected by any open path.

        Does breadth-first search from both X and Y and meets in the middle.
        Paths are walked according to the d-separation rules where paths can
        only traverse motifs <-- v <-- or <-- v --> or --> v --> or
        --> [v] <-- where [.] indicates that v is conditioned on.
        Furthermore, paths nodes (v, t) need to fulfill max_lag <= t <= 0
        and links cannot be traversed backwards.

        Parameters
        ----------
        X, Y : lists of tuples
            Of the form [(var, -tau)], where var specifies the variable
            index and tau the time lag.
        conds : list of tuples
            Of the form [(var, -tau)], where var specifies the variable
            index and tau the time lag.
        max_lag : int
            Maximum time lag.
        starts_with : {None, 'tail', 'arrohead'}
            Whether to only consider paths starting with particular mark at X.
        ends_with : {None, 'tail', 'arrohead'}
            Whether to only consider paths ending with particular mark at Y.
        """
        if max_lag is None:
            if conds is None:
                conds = []
            max_lag = self._get_maximum_possible_lag(X+Y+conds)

        def _walk_to_parents(v, fringe, this_path, other_path):
            """Helper function to update paths when walking to parents."""
            found_connection = False
            for w in self._get_lagged_parents(v, 
                only_non_causal_paths=only_non_causal_paths, X=X, 
                causal_children=causal_children):
                # Cannot walk into conditioned parents and
                # cannot walk beyond t or max_lag
                i, t = w

                if w == x and starts_with == 'arrowhead':
                    continue

                if w == y and ends_with == 'arrowhead':
                    continue

                if (w not in conds and w not in forbidden_nodes and
                    # (w, v) not in seen_links and
                    t <= 0 and abs(t) <= max_lag):
                    # if ((w, 'tail') not in this_path and 
                    #     (w, None) not in this_path):
                    if (w not in this_path or 
                        ('tail' not in this_path[w] and None not in this_path[w])):
                        if self.verbosity > 1:
                            print("Walk parent: %s --> %s  " %(v, w))
                        fringe.append((w, 'tail'))
                        if w not in this_path:
                            this_path[w] = {'tail' : (v, 'arrowhead')}
                        else:
                            this_path[w]['tail'] = (v, 'arrowhead')
                        # seen_links.append((v, w))
                    # Determine whether X and Y are connected
                    # (w, None) indicates the start or end node X/Y
                    # if ((w, 'tail') in other_path 
                    #    or (w, 'arrowhead') in other_path
                    #    or (w, None) in other_path):
                    if w in other_path:
                        found_connection = (w, 'tail') 
                        if self.verbosity > 1:
                            print("Found connection: ", found_connection)  
                        break
            return found_connection, fringe, this_path

        def _walk_to_children(v, fringe, this_path, other_path):
            """Helper function to update paths when walking to children."""
            found_connection = False
            for w in self._get_lagged_children(v, children, 
                only_non_causal_paths=only_non_causal_paths, X=X, 
                causal_children=causal_children):
                # You can also walk into conditioned children,
                # but cannot walk beyond t or max_lag
                i, t = w

                if w == x and starts_with == 'tail':
                    continue

                if w == y and ends_with == 'tail':
                    continue

                if (w not in forbidden_nodes and
                    # (w, v) not in seen_links and
                    t <= 0 and abs(t) <= max_lag):
                    # if ((w, 'arrowhead') not in this_path and 
                    #     (w, None) not in this_path):
                    if (w not in this_path or 
                        ('arrowhead' not in this_path[w] and None not in this_path[w])):
                        if self.verbosity > 1:
                            print("Walk child:  %s --> %s  " %(v, w))
                        fringe.append((w, 'arrowhead'))
                        # this_path[(w, 'arrowhead')] = (v, 'tail')
                        if w not in this_path:
                            this_path[w] = {'arrowhead' : (v, 'tail')}
                        else:
                            this_path[w]['arrowhead'] = (v, 'tail')
                        # seen_links.append((v, w))
                    # Determine whether X and Y are connected
                    # If the other_path contains w with a tail, then w must
                    # NOT be conditioned on. Alternatively, if the other_path
                    # contains w with an arrowhead, then w must be
                    # conditioned on.
                    # if (((w, 'tail') in other_path and w not in conds)
                    #    or ((w, 'arrowhead') in other_path and w in conds)
                    #    or (w, None) in other_path):
                    if w in other_path:
                        if (('tail' in other_path[w] and w not in conds) or
                            ('arrowhead' in other_path[w] and w in conds) or
                            (None in other_path[w])):
                            found_connection = (w, 'arrowhead') 
                            if self.verbosity > 1:
                                print("Found connection: ", found_connection) 
                            break
            return found_connection, fringe, this_path

        def _walk_fringe(this_level, fringe, this_path, other_path):
            """Helper function to walk each fringe, i.e., the path from X and Y,
            respectively."""
            found_connection = False

            if starts_with == 'arrowhead':
                if len(this_level) == 1 and this_level[0] == (x, None):
                    (found_connection, fringe,
                         this_path) = _walk_to_parents(x, fringe, 
                                                       this_path, other_path)
                    return found_connection, fringe, this_path, other_path

            elif starts_with == 'tail':
                if len(this_level) == 1 and this_level[0] == (x, None):
                    (found_connection, fringe,
                         this_path) = _walk_to_children(x, fringe, 
                                                       this_path, other_path)
                    return found_connection, fringe, this_path, other_path 

            if ends_with == 'arrowhead':
                if len(this_level) == 1 and this_level[0] == (y, None):
                    (found_connection, fringe,
                         this_path) = _walk_to_parents(y, fringe, 
                                                       this_path, other_path)
                    return found_connection, fringe, this_path, other_path

            elif ends_with == 'tail':
                if len(this_level) == 1 and this_level[0] == (y, None):
                    (found_connection, fringe,
                         this_path) = _walk_to_children(y, fringe, 
                                                       this_path, other_path)
                    return found_connection, fringe, this_path, other_path 

            for v, mark in this_level:
                if v in conds:
                    if (mark == 'arrowhead' or mark == None) and directed is False:
                        # Motif: --> [v] <--
                        # If standing on a condition and coming from an
                        # arrowhead, you can only walk into parents
                        (found_connection, fringe,
                         this_path) = _walk_to_parents(v, fringe, 
                                                       this_path, other_path)
                        if found_connection: break            
                else:
                    if only_collider_paths_with_vancs:
                        continue

                    if (mark == 'tail' or mark == None):
                        # Motif: <-- v <-- or <-- v -->
                        # If NOT standing on a condition and coming from
                        # a tail mark, you can walk into parents or 
                        # children
                        (found_connection, fringe,
                         this_path) = _walk_to_parents(v, fringe, 
                                                       this_path, other_path)
                        if found_connection: break 
                        
                        if not directed:
                            (found_connection, fringe,
                             this_path) = _walk_to_children(v, fringe, 
                                                           this_path, other_path)
                            if found_connection: break 
                      
                    elif mark == 'arrowhead':
                        # Motif: --> v -->
                        # If NOT standing on a condition and coming from
                        # an arrowhead mark, you can only walk into
                        # children
                        (found_connection, fringe,
                         this_path) = _walk_to_children(v, fringe, 
                                                       this_path, other_path)
                        if found_connection: break

                        if check_optimality_cond and v[0] in self.observed_vars:
                            # if v is not descendant of YM
                            # and v is not connected to Y given X OS\Cu
                            # print("v = ", v)
                            cond4a = v not in optimality_cond_des_YM
                            cond4b = not self._has_any_path(X=[v], Y=optimality_cond_Y, 
                                conds=conds + X, 
                                max_lag=None, 
                                starts_with=None,
                                ends_with=None,  
                                forbidden_nodes=None, #list(prelim_Oset), 
                                return_path=False)
                            # print(cond4a, cond4b)
                            if cond4a and cond4b:
                                (found_connection, fringe,
                                 this_path) = _walk_to_parents(v, fringe, 
                                                               this_path, other_path)
                                # print(found_connection)
                                if found_connection: break

            if self.verbosity > 1:
                print("Updated fringe: ", fringe)
            return found_connection, fringe, this_path, other_path

        def backtrace_path():
            """Helper function to get path from start point, end point, 
            and connection found."""

            path = [found_connection[0]]
            node, mark = found_connection

            if 'tail' in pred[node]:
                mark = 'tail'
            else:
                mark = 'arrowhead'
            # print(found_connection)
            while path[-1] != x:
                # print(path, node, mark, pred[node])
                prev_node, prev_mark = pred[node][mark]
                path.append(prev_node)
                if prev_mark == 'arrowhead':
                    if prev_node not in conds:
                        # if pass_through_colliders:
                        #     if 'tail' in pred[prev_node] and pred[prev_node]['tail'] != (node, mark):
                        #         mark = 'tail'
                        #     else:
                        #         mark = 'arrowhead'
                        # else:
                            mark = 'tail'
                    elif prev_node in conds:
                        mark = 'arrowhead'
                elif prev_mark == 'tail':
                    if 'tail' in pred[prev_node] and pred[prev_node]['tail'] != (node, mark):
                        mark = 'tail'
                    else:
                        mark = 'arrowhead' 
                node = prev_node

            path.reverse()

            node, mark = found_connection
            if 'tail' in succ[node]:
                mark = 'tail'
            else:
                mark = 'arrowhead'

            while path[-1] != y:
                next_node, next_mark = succ[node][mark]
                path.append(next_node)
                if next_mark == 'arrowhead':
                    if next_node not in conds:
                        # if pass_through_colliders:
                        #     if 'tail' in succ[next_node] and succ[next_node]['tail'] != (node, mark):
                        #         mark = 'tail'
                        #     else:
                        #         mark = 'arrowhead'
                        # else:
                            mark = 'tail'
                    elif next_node in conds:
                        mark = 'arrowhead'
                elif next_mark == 'tail':
                    if 'tail' in succ[next_node] and succ[next_node]['tail'] != (node, mark):
                        mark = 'tail'
                    else:
                        mark = 'arrowhead' 
                node = next_node

            return path


        if conds is None:
            conds = []

        if forbidden_nodes is None:
            forbidden_nodes = []

        conds = [z for z in conds if z not in Y and z not in X]
        # print(X, Y, conds)

        if self.selection_vars is not None:
            for selection_var in self.selection_vars:
                conds += [(selection_var, -tau_sel) for tau_sel in range(0, max_lag + 1)]


        N = len(self.links)
        children = self._get_children()

        if only_non_causal_paths:
            anc_Y_dict = self._get_non_blocked_ancestors(Y=Y, conds=None, mode='max_lag',
                                    max_lag=max_lag)[0]
            # print(anc_Y_dict)
            anc_Y = []
            for y in Y:
                anc_Y += anc_Y_dict[y]
            des_X = self._get_descendants(X, children=children, max_lag=max_lag)
            mediators = set(anc_Y).intersection(set(des_X)) - set(Y) - set(X)

            causal_children = list(mediators) + Y
        else:
            causal_children = None

        if only_collider_paths_with_vancs:
            vancs_dict = self._get_non_blocked_ancestors(Y=XYS, conds=None, mode='max_lag',
                                    max_lag=max_lag)[0]
            vancs = set()
            for xys in XYS:
                vancs = vancs.union(set(vancs_dict[xys]))
            vancs = list(vancs) + XYS
            conds = vancs
        # else:
        #     vancs = None

        # Iterate through nodes in X and Y
        for x in X:
          for y in Y:

            # seen_links = []
            # predecessor and successors in search
            # (x, None) where None indicates start/end nodes, later (v,
            # 'tail') or (w, 'arrowhead') indicate how a link ends at a node
            pred = {x : {None: None}}
            succ = {y : {None: None}}

            # initialize fringes, start with forward from X
            forward_fringe = [(x, None)]
            reverse_fringe = [(y, None)]

            while forward_fringe and reverse_fringe:
                if len(forward_fringe) <= len(reverse_fringe):
                    if self.verbosity > 1:
                        print("Walk from X since len(X_fringe)=%d "
                              "<= len(Y_fringe)=%d" % (len(forward_fringe), 
                                len(reverse_fringe)))
                    this_level = forward_fringe
                    forward_fringe = []    
                    (found_connection, forward_fringe, pred, 
                     succ) = _walk_fringe(this_level, forward_fringe, pred, 
                                                succ)

                    # print(pred)
                    if found_connection: 
                        if return_path:
                            backtraced_path = backtrace_path()
                            return [(self.observed_vars.index(node[0]), node[1]) 
                                    for node in backtraced_path 
                                    if node[0] in self.observed_vars]
                        else: 
                            return True
                else:
                    if self.verbosity > 1:
                        print("Walk from Y since len(X_fringe)=%d "
                              "> len(Y_fringe)=%d" % (len(forward_fringe), 
                                len(reverse_fringe)))
                    this_level = reverse_fringe
                    reverse_fringe = []
                    (found_connection, reverse_fringe, succ, 
                     pred) = _walk_fringe(this_level, reverse_fringe, succ, 
                                                pred)

                    if found_connection: 
                        if return_path:
                            backtraced_path = backtrace_path()
                            return [(self.observed_vars.index(node[0]), node[1]) 
                                    for node in backtraced_path 
                                    if node[0] in self.observed_vars]
                        else: 
                            return True

                if self.verbosity > 1:
                    print("X_fringe = %s \n" % str(forward_fringe) +
                          "Y_fringe = %s" % str(reverse_fringe))           

        return False

    def _is_dsep(self, X, Y, Z, max_lag=None):
        """Returns whether X and Y are d-separated given Z in the graph.

        X, Y, Z are of the form (var, lag) for lag <= 0. D-separation is
        based on:

        1. Assessing the maximum time lag max_lag possible for any confounding
        path (see _get_maximum_possible_lag(...)).

        2. Using the time series graph truncated at max_lag we then test
        d-separation between X and Y conditional on Z using breadth-first
        search of non-blocked paths according to d-separation rules.

        Parameters
        ----------
        X, Y, Z : list of tuples
            List of variables chosen for current independence test.
        max_lag : int, optional (default: None)
            Used here to constrain the _is_dsep function to the graph
            truncated at max_lag instead of identifying the max_lag from
            ancestral search.

        Returns
        -------
        dseparated : bool, or path
            True if X and Y are d-separated given Z in the graph.
        """

        N = len(self.links)

        if self.verbosity > 0:
            print("Testing X=%s d-sep Y=%s given Z=%s in TSG" %(X, Y, Z))

        if Z is None:
            Z = []

        if max_lag is not None:
            # max_lags = dict([(j, max_lag) for j in range(N)])
            if self.verbosity > 0:
                print("Set max. time lag to: ", max_lag)
        else:
            max_lag = self._get_maximum_possible_lag(X+Y+Z)

        # Store overall max. lag
        self.max_lag = max_lag

        # _has_any_path is the main function that searches open paths
        any_path = self._has_any_path(X, Y, conds=Z, max_lag=max_lag)

        if any_path:
            dseparated = False
        else:
            dseparated = True

        return dseparated

    def check_shortest_path(self, X, Y, Z,
                 max_lag=None,  # compute_ancestors=False, 
                 starts_with=None, ends_with=None, 
                 forbidden_nodes=None,
                 directed=False,
                 only_non_causal_paths=False,
                 check_optimality_cond=False,
                 optimality_cond_des_YM=None,
                 optimality_cond_Y=None,
                 return_path=False):
        """Returns path between X and Y given Z in the graph.

        X, Y, Z are of the form (var, lag) for lag <= 0. D-separation is
        based on:

        1. Assessing maximum time lag max_lag of last ancestor of any X, Y, Z
        with non-blocked (by Z), non-repeating directed path towards X, Y, Z
        in the graph. 'non_repeating' means that an ancestor X^i_{ t-\tau_i}
        with link X^i_{t-\tau_i} --> X^j_{ t-\tau_j} is only included if
        X^i_{t'-\tau_i} --> X^j_{ t'-\tau_j} for t'!=t is not already part of
        the ancestors.

        2. Using the time series graph truncated at max_lag we then test
        d-separation between X and Y conditional on Z using breadth-first
        search of non-blocked paths according to d-separation rules including
        selection variables.

        Optionally only considers paths starting/ending with specific marks)
        and makes available the ancestors up to max_lag of X, Y, Z. This may take 
        a very long time, however.

        Parameters
        ----------
        X, Y, Z : list of tuples
            List of variables chosen for testing paths.
        max_lag : int, optional (default: None)
            Used here to constrain the has_path function to the graph
            truncated at max_lag instead of identifying the max_lag from
            ancestral search.
        compute_ancestors : bool
            Whether to also make available the ancestors for X, Y, Z as
            self.anc_all_x, self.anc_all_y, and self.anc_all_z, respectively.
        starts_with : {None, 'tail', 'arrohead'}
            Whether to only consider paths starting with particular mark at X.
        ends_with : {None, 'tail', 'arrohead'}
            Whether to only consider paths ending with particular mark at Y.

        Returns
        -------
        path : list or False
            Returns path or False if no path exists.
        """

        N = len(self.links)

        # Translate from observed_vars index to full variable set index
        X = [(self.observed_vars[x[0]], x[1]) for x in X]
        Y = [(self.observed_vars[y[0]], y[1]) for y in Y]
        Z = [(self.observed_vars[z[0]], z[1]) for z in Z]

        # print(X)
        # print(Y)
        # print(Z)

        if check_optimality_cond:
            optimality_cond_des_YM = [(self.observed_vars[x[0]], x[1]) 
                                        for x in optimality_cond_des_YM]
            optimality_cond_Y = [(self.observed_vars[x[0]], x[1]) 
                                    for x in optimality_cond_Y]

        # Get the array to test on
        X, Y, Z = self._check_XYZ(X, Y, Z)

        if self.verbosity > 0:
            print("Testing X=%s d-sep Y=%s given Z=%s in TSG" %(X, Y, Z))

        if max_lag is not None:
            # max_lags = dict([(j, max_lag) for j in range(N)])
            if self.verbosity > 0:
                print("Set max. time lag to: ", max_lag)
        else:
            max_lag = self._get_maximum_possible_lag(X+Y+Z)

        # Store overall max. lag
        self.max_lag = max_lag

        # _has_any_path is the main function that searches open paths
        any_path = self._has_any_path(X, Y, conds=Z, max_lag=max_lag, 
                                      starts_with=starts_with, ends_with=ends_with,
                                      return_path=return_path,
                                      directed=directed,
                                      only_non_causal_paths=only_non_causal_paths,
                                      check_optimality_cond=check_optimality_cond,
                                      optimality_cond_des_YM=optimality_cond_des_YM,
                                      optimality_cond_Y=optimality_cond_Y,
                                      forbidden_nodes=forbidden_nodes)

        if any_path:
            if return_path:
                any_path_observed = [(self.observed_vars.index(node[0]), node[1]) for node in any_path 
                             if node[0] in self.observed_vars]
            else:
                any_path_observed = True
        else: 
            any_path_observed = False

        if self.verbosity > 0:
            print("_has_any_path     = ", any_path)
            print("_has_any_path_obs = ", any_path_observed)


        # if compute_ancestors:
        #     if self.verbosity > 0:
        #         print("Compute ancestors.")

        #     # Get ancestors up to maximum ancestral time lag incl. repeated
        #     # links
        #     self.anc_all_x, _ = self._get_non_blocked_ancestors(X, conds=Z,
        #                                     mode='max_lag', max_lag=max_lag)
        #     self.anc_all_y, _ = self._get_non_blocked_ancestors(Y, conds=Z,
        #                                     mode='max_lag', max_lag=max_lag)
        #     self.anc_all_z, _ = self._get_non_blocked_ancestors(Z, conds=Z,
        #                                     mode='max_lag', max_lag=max_lag)

        return any_path_observed

    def run_test(self, X, Y, Z=None, tau_max=0, cut_off='2xtau_max', alpha_or_thres=None,
                 verbosity=0):
        """Perform oracle conditional independence test.

        Calls the d-separation function.

        Parameters
        ----------
        X, Y, Z : list of tuples
            X,Y,Z are of the form [(var, -tau)], where var specifies the
            variable index in the observed_vars and tau the time lag.
        tau_max : int, optional (default: 0)
            Not used here.
        cut_off : {'2xtau_max', 'max_lag', 'max_lag_or_tau_max'}
            Not used here.
        alpha_or_thres : float
            Not used here.

        Returns
        -------
        val, pval : Tuple of floats
            The test statistic value and the p-value.
        """

        if Z is None:
            Z = []

        # Translate from observed_vars index to full variable set index
        X = [(self.observed_vars[x[0]], x[1]) for x in X]
        Y = [(self.observed_vars[y[0]], y[1]) for y in Y]
        Z = [(self.observed_vars[z[0]], z[1]) for z in Z]

        # Get the array to test on
        X, Y, Z = self._check_XYZ(X, Y, Z)

        if not str((X, Y, Z)) in self.dsepsets:
            self.dsepsets[str((X, Y, Z))] = self._is_dsep(X, Y, Z)

        if self.dsepsets[str((X, Y, Z))]:
            val = 0.
            pval = 1.
            dependent = False
        else:
            val = 1.
            pval = 0.
            dependent = True

        # Saved here, but not currently used
        self.ci_results[(tuple(X), tuple(Y),tuple(Z))] = (val, pval, dependent)
        
        if verbosity > 1:
            self._print_cond_ind_results(val=val, pval=pval, cached=False,
                                         conf=None)
        # Return the value and the pvalue
        if alpha_or_thres is None:
            return val, pval
        else:
            return val, pval, dependent

    def get_measure(self, X, Y, Z=None, tau_max=0):
        """Returns dependence measure.

        Returns 0 if X and Y are d-separated given Z in the graph and 1 else.

        Parameters
        ----------
        X, Y [, Z] : list of tuples
            X,Y,Z are of the form [(var, -tau)], where var specifies the
            variable index in the observed_vars and tau the time lag.

        tau_max : int, optional (default: 0)
            Maximum time lag. This may be used to make sure that estimates for
            different lags in X, Z, all have the same sample size.

        Returns
        -------
        val : float
            The test statistic value.

        """

        # Translate from observed_vars index to full variable set index
        X = [(self.observed_vars[x[0]], x[1]) for x in X]
        Y = [(self.observed_vars[y[0]], y[1]) for y in Y]
        Z = [(self.observed_vars[z[0]], z[1]) for z in Z]

        # Check XYZ
        X, Y, Z = _check_XYZ(X, Y, Z)

        if not str((X, Y, Z)) in self.dsepsets:
            self.dsepsets[str((X, Y, Z))] = self._is_dsep(X, Y, Z)

        if self.dsepsets[str((X, Y, Z))]:
            return 0.
        else:
            return 1.

    def _print_cond_ind_results(self, val, pval=None, cached=None, conf=None):
        """Print results from conditional independence test.

        Parameters
        ----------
        val : float
            Test stastistic value.
        pval : float, optional (default: None)
            p-value
        conf : tuple of floats, optional (default: None)
            Confidence bounds.
        """
        printstr = "        val = %.3f" % (val)      
        if pval is not None:
            printstr += " | pval = %.5f" % (pval)
        if conf is not None:
            printstr += " | conf bounds = (%.3f, %.3f)" % (
                conf[0], conf[1])
        if cached is not None:
            printstr += " %s" % ({0:"", 1:"[cached]"}[cached])

        print(printstr)

    def get_model_selection_criterion(self, j, parents, tau_max=0):
        """
        Base class assumption that this is not implemented.  Concrete classes
        should override when possible.
        """
        raise NotImplementedError("Model selection not"+\
                                  " implemented for %s" % self.measure)

    def _reverse_patt(self, patt):
        """Inverts a link pattern"""

        if patt == "":
            return ""

        left_mark, middle_mark, right_mark = patt[0], patt[1], patt[2]
        if left_mark == "<":
            new_right_mark = ">"
        else:
            new_right_mark = left_mark
        if right_mark == ">":
            new_left_mark = "<"
        else:
            new_left_mark = right_mark

        return new_left_mark + middle_mark + new_right_mark


    def get_links_from_graph(self, graph):
        """
        Constructs links_coeffs dictionary, observed_vars, 
        and selection_vars from graph array (MAG or DAG).

        In the case of MAGs, for every <-> or --- link further
        latent and selection variables, respectively, are added.
        This corresponds to a canonical DAG (Richardson Spirtes 2002).

        For ADMGs "---" are not supported, but also links of type "+->"
        exist, which corresponds to having both "-->" and "<->".

        Can be used to evaluate d-separation in MAG/DAGs.

        """

        if "U3" not in str(graph.dtype):
            raise ValueError("graph must be of type '<U3'!")

        if self.graph_is_mag:
            edge_types = ["-->", "<--", "<->", "---"]
        else:
            edge_types = ["-->", "<--", "<->", "+->", "<-+"] #, "--+", "+--"]


        N, N, tau_maxplusone = graph.shape
        tau_max = tau_maxplusone - 1

        observed_vars = list(range(N))

        selection_vars = []

        links = {j: [] for j in observed_vars }

        # Add further latent variables to accommodate <-> and --- links
        latent_index = N
        for i, j, tau in zip(*np.where(graph)):

            edge_type = graph[i, j, tau]

            if edge_type not in edge_types:
                raise ValueError(
                    "Links can only be in %s " %str(edge_types)
                )

            if tau == 0:
                if edge_type != self._reverse_patt(graph[j, i, 0]):
                    raise ValueError(
                        "graph needs to have consistent lag-zero patterns (eg"
                        " graph[i,j,0]='-->' requires graph[j,i,0]='<--')"
                    )

                # Consider contemporaneous links only once
                if j > i:
                    continue

            # Restrict lagged links
            else: 
                if edge_type not in ["-->", "<->", "---", "+->"]: #, "--+"]:
                    raise ValueError(
                        "Lagged links can only be in ['-->', '<->', '---', '+->']"
                    )

            if edge_type == "-->":
                links[j].append((i, -tau))
            elif edge_type == "<--":
                links[i].append((j, -tau))
            elif edge_type == "<->":
                links[latent_index] = []
                links[i].append((latent_index, 0))
                links[j].append((latent_index, -tau))
                latent_index += 1
            elif edge_type == "---":
                links[latent_index] = []
                selection_vars.append(latent_index)
                links[latent_index].append((i, -tau))
                links[latent_index].append((j, 0))
                latent_index += 1
            elif edge_type == "+->":
                links[j].append((i, -tau))
                links[latent_index] = []
                links[i].append((latent_index, 0))
                links[j].append((latent_index, -tau))
                latent_index += 1
            elif edge_type == "<-+":
                links[i].append((j, -tau))
                links[latent_index] = []
                links[i].append((latent_index, 0))
                links[j].append((latent_index, -tau))
                latent_index += 1
            # elif edge_type == "+--":
            #     links[i].append((j, -tau))
            #     links[latent_index] = []
            #     selection_vars.append(latent_index)
            #     links[latent_index].append((i, -tau))
            #     links[latent_index].append((j, 0))
            #     latent_index += 1
            # elif edge_type == "--+":
            #     links[j].append((i, -tau))
            #     links[latent_index] = []
            #     selection_vars.append(latent_index)
            #     links[latent_index].append((i, -tau))
            #     links[latent_index].append((j, 0))
            #     latent_index += 1

        return links, observed_vars, selection_vars

    def _get_minmax_lag(self, links):
        """Helper function to retrieve tau_min and tau_max from links
        """

        N = len(links)

        # Get maximum time lag
        min_lag = np.inf
        max_lag = 0
        for j in range(N):
            for link_props in links[j]:
                if len(link_props) == 3:
                    i, lag = link_props[0]
                    coeff = link_props[1]
                else:
                    i, lag = link_props
                    coeff = 1.                
                # func = link_props[2]
                if coeff != 0.:
                    min_lag = min(min_lag, abs(lag))
                    max_lag = max(max_lag, abs(lag))
        return min_lag, max_lag

    def get_graph_from_links(self, tau_max=None):
        """
        Constructs graph (DAG or MAG or ADMG) from links, observed_vars, 
        and selection_vars.

        For ADMGs uses the Latent projection operation (Pearl 2009).

        """

        # TODO: use MAG from DAG construction procedure (lecture notes)
        # issues with tau_max?
        if self.graph_is_mag is False and len(self.selection_vars) > 0:
            raise ValueError("ADMGs do not support selection_vars.")

        N_all = len(self.links)

        # If tau_max is None, compute from links_coeffs
        _, max_lag_links = self._get_minmax_lag(self.links)
        if tau_max is None:
            tau_max = max_lag_links
        else:
            if max_lag_links > tau_max:
                raise ValueError("tau_max must be >= maximum lag in links_coeffs; choose tau_max=None")

        # print("max_lag_links ", max_lag_links)

        N = len(self.observed_vars)

        # Init graph
        graph = np.zeros((N, N, tau_max + 1), dtype='<U3')
        graph[:] = ""
        # We will enumerate the observed variables with (i,j) which refers to the index in MAG graph
        # while x, y iterates through the variables in the underlying DAG

        # Loop over the observed variables
        for j, y in enumerate(self.observed_vars):
          for i, x in enumerate(self.observed_vars):
            for tau in range(0, tau_max + 1):
              if (x, -tau) != (y, 0):

                if self.graph_is_mag:
                    dag_anc_y, _ = self._get_non_blocked_ancestors(Y=[(y, 0)], conds=None, 
                                            mode='max_lag',
                                            max_lag=tau_max)
                    # Only consider observed ancestors
                    mag_anc_y = [anc for anc in dag_anc_y[(y, 0)] 
                                if anc[0] in self.observed_vars]

                    dag_anc_x, _ = self._get_non_blocked_ancestors(Y=[(x, -tau)], 
                                    conds=None, mode='max_lag',
                                    max_lag=tau_max)
                    
                    # Only consider observed ancestors
                    mag_anc_x = [anc for anc in dag_anc_x[(x, -tau)] 
                                if anc[0] in self.observed_vars]   

                    # Add selection variable ancestors
                    dag_anc_s = set()
                    for s in self.selection_vars:
                        dag_anc_s_here, _ = self._get_non_blocked_ancestors(Y=[(s, 0)], 
                                    conds=None, mode='max_lag',
                                    max_lag=tau_max)
                        dag_anc_s = dag_anc_s.union(set(dag_anc_s_here[(s, 0)]))
                    
                    dag_anc_s = list(dag_anc_s)
                    # Only consider observed ancestors
                    mag_anc_s = [anc for anc in dag_anc_s 
                                if anc[0] in self.observed_vars]

                    Z = set([z for z in mag_anc_y + mag_anc_x + mag_anc_s if z != (y, 0) and z != (x, -tau)])
                    Z = list(Z)
                    
                    separated = self._is_dsep(X=[(x, -tau)], Y=[(y, 0)], Z=Z, max_lag=None)
                    
                    # If X and Y are connected given Z, mark a link
                    if not separated:
                        # (i, -tau) --> j
                        if (x, -tau) in dag_anc_y[(y, 0)] + dag_anc_s and (y, 0) not in dag_anc_x[(x, -tau)] + dag_anc_s:
                            graph[i, j, tau] = "-->"
                            if tau == 0:
                                graph[j, i, 0] = "<--"

                        elif (x, -tau) not in dag_anc_y[(y, 0)] + dag_anc_s and (y, 0) not in dag_anc_x[(x, -tau)] + dag_anc_s:
                            graph[i, j, tau] = "<->"
                            if tau == 0:
                                graph[j, i, 0] = "<->"

                        elif (x, -tau) in dag_anc_y[(y, 0)] + dag_anc_s and (y, 0) in dag_anc_x[(x, -tau)] + dag_anc_s:
                            graph[i, j, tau] = "---"
                            if tau == 0:
                                graph[j, i, 0] = "---"
                else:
                    if tau == 0 and j >= i:
                        continue
                    # edge_types = ["-->", "<->", "+->"]
                    # Latent projection operation:
                    # (i)  ADMG contains i --> j iff there is a directed path x --> ... --> y on which
                    #      every non-endpoint vertex is in hidden variables (= not in observed_vars)
                    # (ii) ADMG contains i <-> j iff there exists a path of the form x <-- ... --> y on
                    #      which every non-endpoint vertex is non-collider AND in L (=not in observed_vars)
                    observed_varslags = set([(v, -lag) for v in self.observed_vars
                                                   for lag in range(0, tau_max + 1)]) - set([(x, -tau), (y, 0)])
                    cond_one_xy = self._has_any_path(X=[(x, -tau)], Y=[(y, 0)], 
                        conds=[], 
                        max_lag=None, 
                        starts_with='tail',
                        ends_with='arrowhead',
                        directed=True,  
                        forbidden_nodes=list(observed_varslags), 
                        return_path=False)
                    if tau == 0:
                        cond_one_yx = self._has_any_path(X=[(y, 0)], Y=[(x, 0)], 
                            conds=[], 
                            max_lag=None, 
                            starts_with='tail',
                            ends_with='arrowhead',
                            directed=True,  
                            forbidden_nodes=list(observed_varslags), 
                            return_path=False)
                    else:
                        cond_one_yx = False
                    cond_two = self._has_any_path(X=[(x, -tau)], Y=[(y, 0)], 
                        conds=[], 
                        max_lag=None, 
                        starts_with='arrowhead',
                        ends_with='arrowhead',
                        directed=False,  
                        forbidden_nodes=list(observed_varslags), 
                        return_path=False)
                    if cond_one_xy and cond_one_yx:
                        raise ValueError("Cyclic graph!")
                    # print((x, -tau), y, cond_one_xy, cond_one_yx, cond_two)

                    # Only (i) holds: i --> j
                    if cond_one_xy and not cond_two:
                        graph[i, j, tau] = "-->"
                        if tau == 0:
                            graph[j, i, 0] = "<--"
                    elif cond_one_yx and not cond_two:
                        graph[i, j, tau] = "<--"
                        if tau == 0:
                            graph[j, i, 0] = "-->"

                    # Only (ii) holds: i <-> j
                    elif not cond_one_xy and not cond_one_yx and cond_two:
                        graph[i, j, tau] = "<->"
                        if tau == 0:
                            graph[j, i, 0] = "<->"

                    # Both (i) and (ii) hold: i +-> j
                    elif cond_one_xy and cond_two:
                        graph[i, j, tau] = "+->"
                        if tau == 0:
                            graph[j, i, 0] = "<-+"
                    elif cond_one_yx and cond_two:
                        graph[i, j, tau] = "<-+"
                        if tau == 0:
                            graph[j, i, 0] = "+->"
                    # print((i, -tau), j, cond_one_xy, cond_one_yx, cond_two)

        return graph

    def get_confidence(self, X, Y, Z=None, tau_max=0):
        """For compatibility with PCMCI.

        Returns
        -------
        None
        """
        return None

if __name__ == '__main__':

    import tigramite.plotting as tp
    from matplotlib import pyplot as plt
    def lin_f(x): return x

    # Define the stationary DAG
    links = {0 : [(0, -3), (1, 0)], 1: [(2, -2)], 2: [(1, -2)]}
    observed_vars = [0, 1, 2]

    oracle = OracleCI(links=links, 
        observed_vars=observed_vars, 
        graph_is_mag=True,
        #         selection_vars=selection_vars,
        #     verbosity=2
        )
    graph = oracle.graph
    print(graph[:,:,0])

    tp.plot_time_series_graph(graph=graph, var_names=None, figsize=(5, 5),
                save_name="tsg.pdf")

    X = [(0, 0)]
    Y = [(2, 0)]
    Z = []
    # node = (3, 0)
    # prelim_Oset = set([(3, 0)])
    # S = set([])
    # collider_path_nodes = set([])
    path = oracle._has_any_path(X=X, Y=Y, 
                            conds=Z, 
                            max_lag=8, 
                            starts_with='arrowhead',
                            ends_with='arrowhead',  
                            forbidden_nodes=None, 
                            return_path=True)
    print(path)

    print("-------------------------------")
    print(oracle._get_maximum_possible_lag(X+Z)) #(X = X, Y = Y, Z = Z))

    # cond_ind_test = OracleCI(graph=graph)
    # links, observed_vars, selection_vars = cond_ind_test.get_links_from_graph(graph)
    # print("{")
    # for j in links.keys():
    #     parents = repr([(p, 'coeff', 'lin_f') for p in links[j]])
    #     print(f"{j: 1d}" ":"  f"{parents:s},")
    # print(repr(observed_vars))
    # cond_ind_test = OracleCI(graph=graph, verbosity=2)

    # X = [(0, 0)]
    # Y = [(2, 0)]
    # Z = [(7, 0), (3, 0), (6, 0), (5, 0), (4, 0)] #(1, -3), (1, -2), (0, -2), (0, -1), (0, -3)]
    # #(j, -2) for j in range(N)] + [(j, 0) for j in range(N)]

    # # print(oracle._get_non_blocked_ancestors(Z, Z=None, mode='max_lag',
    # #                                     max_lag=2))
    # # cond_ind_test = OracleCI(links, observed_vars=observed_vars, verbosity=2)

    # print(cond_ind_test.get_shortest_path(X=X, Y=Y, Z=Z,
    #              max_lag=None, compute_ancestors=False, 
    #              backdoor=True))
   
    # anc_x=None  #oracle.anc_all_x[X[0]]
    # anc_y=None #oracle.anc_all_y[Y[0]]
    # anc_xy=None # []
    # # # for z in Z:
    # # #     anc_xy += oracle.anc_all_z[z]
    
    # fig, ax = tp.plot_tsg(links, 
    #             X=[(observed_vars[x[0]], x[1]) for x in X], 
    #             Y=[(observed_vars[y[0]], y[1]) for y in Y], 
    #             Z=[(observed_vars[z[0]], z[1]) for z in Z],
    #     anc_x=anc_x, anc_y=anc_y, 
    #     anc_xy=anc_xy)

    # fig.savefig("/home/rung_ja/Downloads/tsg.pdf")



"""Tigramite causal discovery for time series."""

# Author: Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

from __future__ import print_function
import warnings
import itertools
from collections import defaultdict
from copy import deepcopy
import numpy as np
import scipy.stats
import math
from joblib import Parallel, delayed

class PCMCIbase():
    r"""PCMCI base class.

    Parameters
    ----------
    dataframe : data object
        This is the Tigramite dataframe object. Among others, it has the
        attributes dataframe.values yielding a numpy array of shape (
        observations T, variables N) and optionally a mask of the same shape.
    cond_ind_test : conditional independence test object
        This can be ParCorr or other classes from
        ``tigramite.independence_tests`` or an external test passed as a
        callable. This test can be based on the class
        tigramite.independence_tests.CondIndTest.
    verbosity : int, optional (default: 0)
        Verbose levels 0, 1, ...

    Attributes
    ----------
    all_parents : dictionary
        Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing
        the conditioning-parents estimated with PC algorithm.
    val_min : dictionary
        Dictionary of form val_min[j][(i, -tau)] = float
        containing the minimum test statistic value for each link estimated in
        the PC algorithm.
    pval_max : dictionary
        Dictionary of form pval_max[j][(i, -tau)] = float containing the maximum
        p-value for each link estimated in the PC algorithm.
    iterations : dictionary
        Dictionary containing further information on algorithm steps.
    N : int
        Number of variables.
    T : dict
        Time series sample length of dataset(s).
    """

    def __init__(self, dataframe,
                 cond_ind_test,
                 verbosity=0):
        # Set the data for this iteration of the algorithm
        self.dataframe = dataframe
        # Set the conditional independence test to be used
        self.cond_ind_test = deepcopy(cond_ind_test)
        if isinstance(self.cond_ind_test, type):
            raise ValueError("PCMCI requires that cond_ind_test "
                             "is instantiated, e.g. cond_ind_test =  "
                             "ParCorr().")
        self.cond_ind_test.set_dataframe(self.dataframe)
        # Set the verbosity for debugging/logging messages
        self.verbosity = verbosity
        # Set the variable names 
        self.var_names = self.dataframe.var_names

        # Store the shape of the data in the T and N variables
        self.T = self.dataframe.T
        self.N = self.dataframe.N


    def _reverse_link(self, link):
        """Reverse a given link, taking care to replace > with < and vice versa."""

        if link == "":
            return ""

        if link[2] == ">":
            left_mark = "<"
        else:
            left_mark = link[2]

        if link[0] == "<":
            right_mark = ">"
        else:
            right_mark = link[0]

        return left_mark + link[1] + right_mark

    def _check_cyclic(self, link_dict):
        """Return True if the link_dict has a contemporaneous cycle.

        """

        path = set()
        visited = set()

        def visit(vertex):
            if vertex in visited:
                return False
            visited.add(vertex)
            path.add(vertex)
            for itaui in link_dict.get(vertex, ()):
                i, taui = itaui
                link_type = link_dict[vertex][itaui] 
                if taui == 0 and link_type in ['-->', '-?>']:
                    if i in path or visit(i):
                        return True
            path.remove(vertex)
            return False

        return any(visit(v) for v in link_dict)

    def _set_link_assumptions(self, link_assumptions, tau_min, tau_max,
                       remove_contemp=False):
        """Helper function to set and check the link_assumptions argument

        Parameters
        ----------
        link_assumptions : dict
            Dictionary of form {j:{(i, -tau): link_type, ...}, ...} specifying
            assumptions about links. This initializes the graph with entries
            graph[i,j,tau] = link_type. For example, graph[i,j,0] = '-->'
            implies that a directed link from i to j at lag 0 must exist.
            Valid link types are 'o-o', '-->', '<--'. In addition, the middle
            mark can be '?' instead of '-'. Then '-?>' implies that this link
            may not exist, but if it exists, its orientation is '-->'. Link
            assumptions need to be consistent, i.e., graph[i,j,0] = '-->'
            requires graph[j,i,0] = '<--' and acyclicity must hold. If a link
            does not appear in the dictionary, it is assumed absent. That is,
            if link_assumptions is not None, then all links have to be specified
            or the links are assumed absent.
        tau_mix : int
            Minimum time delay to test.
        tau_max : int
            Maximum time delay to test.
        remove_contemp : bool
            Whether contemporaneous links (at lag zero) should be removed.

        Returns
        -------
        link_assumptions : dict
            Cleaned links.
        """
        # Copy and pass into the function
        _int_link_assumptions = deepcopy(link_assumptions)
        # Set the default selected links if none are set
        _vars = list(range(self.N))
        _lags = list(range(-(tau_max), -tau_min + 1, 1))
        if _int_link_assumptions is None:
            _int_link_assumptions = {}
            # Set the default as all combinations
            for j in _vars:
                _int_link_assumptions[j] = {}
                for i in _vars:
                    for lag in range(tau_min, tau_max + 1):
                        if not (i == j and lag == 0):
                            if lag == 0:
                                _int_link_assumptions[j][(i, 0)] = 'o?o'
                            else:
                                _int_link_assumptions[j][(i, -lag)] = '-?>'
  
        else:

            if remove_contemp:
                for j in _int_link_assumptions.keys():
                    _int_link_assumptions[j] = {link:_int_link_assumptions[j][link] 
                                        for link in _int_link_assumptions[j]
                                         if link[1] != 0}

        # Make contemporaneous assumptions consistent and orient lagged links
        for j in _vars:
            for link in _int_link_assumptions[j]:
                i, tau = link
                link_type = _int_link_assumptions[j][link]
                if tau == 0:
                    if (j, 0) in _int_link_assumptions[i]:
                        if _int_link_assumptions[j][link] != self._reverse_link(_int_link_assumptions[i][(j, 0)]):
                            raise ValueError("Inconsistent link assumptions for indices %d - %d " %(i, j))
                    else:
                        _int_link_assumptions[i][(j, 0)] = self._reverse_link(_int_link_assumptions[j][link])
                else:
                    # Orient lagged links by time order while leaving the middle mark
                    new_link_type = '-' + link_type[1] + '>'
                    _int_link_assumptions[j][link] = new_link_type

        # Otherwise, check that our assumpions are sane
        # Check that the link_assumptions refer to links that are inside the
        # data range and types
        _key_set = set(_int_link_assumptions.keys())
        valid_entries = _key_set == set(range(self.N))

        valid_types = [
                    'o-o',
                    'o?o',
                    '-->',
                    '-?>',
                    '<--',
                    '<?-',
                        ]

        for links in _int_link_assumptions.values():
            if isinstance(links, dict) and len(links) == 0:
                continue
            for var, lag in links:
                if var not in _vars or lag not in _lags:
                    valid_entries = False
                if links[(var, lag)] not in valid_types:
                    valid_entries = False


        if not valid_entries:
            raise ValueError("link_assumptions"
                             " must be dictionary with keys for all [0,...,N-1]"
                             " variables and contain only links from "
                             "these variables in range [tau_min, tau_max] "
                             "and with link types in %s" %str(valid_types))

        # Check for contemporaneous cycles
        if self._check_cyclic(_int_link_assumptions):
            raise ValueError("link_assumptions has contemporaneous cycle(s).")

        # Return the _int_link_assumptions
        return _int_link_assumptions

    def _dict_to_matrix(self, val_dict, tau_max, n_vars, default=1):
        """Helper function to convert dictionary to matrix format.

        Parameters
        ---------
        val_dict : dict
            Dictionary of form {0:{(0, -1):float, ...}, 1:{...}, ...}.
        tau_max : int
            Maximum lag.
        n_vars : int
            Number of variables.
        default : int
            Default value for entries not part of val_dict.

        Returns
        -------
        matrix : array of shape (N, N, tau_max+1)
            Matrix format of p-values and test statistic values.
        """
        matrix = np.ones((n_vars, n_vars, tau_max + 1))
        matrix *= default

        for j in val_dict.keys():
            for link in val_dict[j].keys():
                k, tau = link
                if tau == 0:
                    matrix[k, j, 0] = matrix[j, k, 0] = val_dict[j][link]
                else:
                    matrix[k, j, abs(tau)] = val_dict[j][link]
        return matrix


    def get_corrected_pvalues(self, p_matrix,
                              fdr_method='fdr_bh',
                              exclude_contemporaneous=True,
                              tau_min=0,
                              tau_max=1,
                              link_assumptions=None,
                              ):
        """Returns p-values corrected for multiple testing.

        Currently implemented is Benjamini-Hochberg False Discovery Rate
        method. Correction is performed either among all links if
        exclude_contemporaneous==False, or only among lagged links.

        Parameters
        ----------
        p_matrix : array-like
            Matrix of p-values. Must be of shape (N, N, tau_max + 1).
        tau_min : int, default: 0
            Minimum time lag. Only used as consistency check of link_assumptions. 
        tau_max : int, default: 1
            Maximum time lag. Must be larger or equal to tau_min. Only used as 
            consistency check of link_assumptions. 
        link_assumptions : dict or None
            Dictionary of form {j:{(i, -tau): link_type, ...}, ...} specifying
            assumptions about links. This initializes the graph with entries
            graph[i,j,tau] = link_type. For example, graph[i,j,0] = '-->'
            implies that a directed link from i to j at lag 0 must exist.
            Valid link types are 'o-o', '-->', '<--'. In addition, the middle
            mark can be '?' instead of '-'. Then '-?>' implies that this link
            may not exist, but if it exists, its orientation is '-->'. Link
            assumptions need to be consistent, i.e., graph[i,j,0] = '-->'
            requires graph[j,i,0] = '<--' and acyclicity must hold. If a link
            does not appear in the dictionary, it is assumed absent. That is,
            if link_assumptions is not None, then all links have to be specified
            or the links are assumed absent.
        fdr_method : str, optional (default: 'fdr_bh')
            Correction method, currently implemented is Benjamini-Hochberg
            False Discovery Rate method.     
        exclude_contemporaneous : bool, optional (default: True)
            Whether to include contemporaneous links in correction.

        Returns
        -------
        q_matrix : array-like
            Matrix of shape (N, N, tau_max + 1) containing corrected p-values.
        """

        def _ecdf(x):
            """No frills empirical cdf used in fdr correction.
            """
            nobs = len(x)
            return np.arange(1, nobs + 1) / float(nobs)

        # Get the shape parameters from the p_matrix
        _, N, tau_max_plusone = p_matrix.shape
        # Check the limits on tau
        self._check_tau_limits(tau_min, tau_max)
        # Include only link_assumptions if given
        if link_assumptions != None:
            # Create a mask for these values
            mask = np.zeros((N, N, tau_max_plusone), dtype='bool')
            _int_link_assumptions = self._set_link_assumptions(link_assumptions, tau_min, tau_max)
            for j, links_ in _int_link_assumptions.items():
                for link in links_:
                    i, lag = link
                    if _int_link_assumptions[j][link] not in ["<--", "<?-"]:    
                        mask[i, j, abs(lag)] = True
        else:
            # Create a mask for these values
            mask = np.ones((N, N, tau_max_plusone), dtype='bool')
        # Ignore values from lag-zero 'autocorrelation' indices
        mask[range(N), range(N), 0] = False
        # Exclude all contemporaneous values if requested
        if exclude_contemporaneous:
            mask[:, :, 0] = False
        # Create the return value
        q_matrix = np.array(p_matrix)
        # Use the multiple tests function
        if fdr_method is None or fdr_method == 'none':
            pass
        elif fdr_method == 'fdr_bh':
            pvs = p_matrix[mask]
            pvals_sortind = np.argsort(pvs)
            pvals_sorted = np.take(pvs, pvals_sortind)

            ecdffactor = _ecdf(pvals_sorted)

            pvals_corrected_raw = pvals_sorted / ecdffactor
            pvals_corrected = np.minimum.accumulate(
                pvals_corrected_raw[::-1])[::-1]
            del pvals_corrected_raw

            pvals_corrected[pvals_corrected > 1] = 1
            pvals_corrected_ = np.empty_like(pvals_corrected)
            pvals_corrected_[pvals_sortind] = pvals_corrected
            del pvals_corrected

            q_matrix[mask] = pvals_corrected_

        else:
            raise ValueError('Only FDR method fdr_bh implemented')

        # Return the new matrix
        return q_matrix


    def _get_adj_time_series(self, graph, include_conflicts=True, sort_by=None):
        """Helper function that returns dictionary of adjacencies from graph.

        Parameters
        ----------
        graph : array of shape [N, N, tau_max+1]
            Resulting causal graph, see description above for interpretation.
        include_conflicts : bool, optional (default: True)
            Whether conflicting links (marked as 2 in graph) should be returned.
        sort_by : dict or none, optional (default: None)
            If not None, the adjacencies are sorted by the absolute values of
            the corresponding entries.

        Returns
        -------
        adjt : dictionary
            Adjacency dictionary.
        """
        N, N, tau_max_plusone = graph.shape
        adjt = {}
        if include_conflicts:
            for j in range(N):
                where = np.where(graph[:, j, :] != "")
                adjt[j] = list(zip(*(where[0], -where[1])))
        else:
            for j in range(N):
                where = np.where(np.logical_and.reduce((graph[:,j,:] != "", 
                                                        graph[:,j,:] != "x-x",
                                                        graph[:,j,:] != "x?x")))
                # where = np.where(graph[:, j, :] == 1)
                adjt[j] = list(zip(*(where[0], -where[1])))

        if sort_by is not None:
            for j in range(N):
                # Get the absolute value for all the test statistics
                abs_values = {k: np.abs(sort_by[j][k]) for k in list(sort_by[j])
                              if k in adjt[j]}
                adjt[j] = sorted(abs_values, key=abs_values.get, reverse=True)

        return adjt

    def _get_adj_time_series_contemp(self, graph, include_conflicts=True,
                                     sort_by=None):
        """Helper function that returns dictionary of contemporaneous
        adjacencies from graph.

        Parameters
        ----------
        graph : array of shape [N, N, tau_max+1]
            Resulting causal graph, see description above for interpretation.
        include_conflicts : bool, optional (default: True)
            Whether conflicting links (marked as 2 in graph) should be returned.
        sort_by : dict or none, optional (default: None)
            If not None, the adjacencies are sorted by the absolute values of
            the corresponding entries.

        Returns
        -------
        adjt : dictionary
            Contemporaneous adjacency dictionary.
        """
        N, N, tau_max_plusone = graph.shape
        adjt = self._get_adj_time_series(graph,
                                         include_conflicts=include_conflicts,
                                         sort_by=sort_by)
        for j in range(N):
            adjt[j] = [a for a in adjt[j] if a[1] == 0]
            # adjt[j] = list(np.where(graph[:,j,0] != 0)[0])

        return adjt


    def _get_simplicial_node(self, circle_cpdag, variable_order):
        """Find simplicial nodes in circle component CPDAG.

        A vertex V is simplicial if all vertices adjacent to V are also adjacent
        to each other (form a clique).

        Parameters
        ----------
        circle_cpdag : array of shape (N, N, tau_max+1)
            Circle component of PCMCIplus graph.
        variable_order : list of length N
            Order of variables in which to search for simplicial nodes.

        Returns
        -------
        (j, adj_j) or None
            First found simplicial node and its adjacencies.
        """

        for j in variable_order:
            adj_j = np.where(np.logical_or(circle_cpdag[:,j,0] == "o-o",
                                           circle_cpdag[:,j,0] == "o?o"))[0].tolist()

            # Make sure the node has any adjacencies
            all_adjacent = len(adj_j) > 0

            # If it has just one adjacency, it's also simplicial
            if len(adj_j) == 1:
                return (j, adj_j)  
            else:
                for (var1, var2) in itertools.combinations(adj_j, 2):
                    if circle_cpdag[var1, var2, 0] == "": 
                        all_adjacent = False
                        break

                if all_adjacent:
                    return (j, adj_j)

        return None

    def _get_dag_from_cpdag(self, cpdag_graph, variable_order):
        """Yields one member of the Markov equivalence class of a CPDAG.

        Removes conflicting edges.

        Used in PCMCI to run model selection on the output of PCMCIplus in order
        to, e.g., optimize pc_alpha.

        Based on Zhang 2008, Theorem 2 (simplified for CPDAGs): Let H be the
        graph resulting from the following procedure applied to a CPDAG:
 
        Consider the circle component of the CPDAG (sub graph consisting of all
        (o-o edges, i.e., only for contemporaneous links), CPDAG^C and turn into
        a DAG with no unshielded colliders. Then (H is a member of the Markov
        equivalence class of the CPDAG.

        We use the approach mentioned in Colombo and Maathuis (2015) Lemma 7.6:
        First note that CPDAG^C is chordal, that is, any cycle of length four or
        more has a chord, which is an edge joining two vertices that are not
        adjacent in the cycle; see the proof of Lemma 4.1 of Zhang (2008b). Any
        chordal graph with more than one vertex has two simplicial vertices,
        that is, vertices V such that all vertices adjacent to V are also
        adjacent to each other. We choose such a vertex V1 and orient any edges
        incident to V1 into V1. Since V1 is simplicial, this does not create
        unshielded colliders. We then remove V1 and these edges from the graph.
        The resulting graph is again chordal and therefore again has at least
        two simplicial vertices. Choose such a vertex V2 , and orient any edges
        incident to V2 into V2. We continue this procedure until all edges are
        oriented. The resulting ordering is called a perfect elimination scheme
        for CPDAG^C. Then the combined graph with the directed edges already
        contained in the CPDAG is returned.

        Parameters
        ----------
        cpdag_graph : array of shape (N, N, tau_max+1)
            Result of PCMCIplus, a CPDAG.
        variable_order : list of length N
            Order of variables in which to search for simplicial nodes.

        Returns
        -------
        dag : array of shape (N, N, tau_max+1)
            One member of the Markov equivalence class of the CPDAG.
        """

        # TODO: Check whether CPDAG is chordal

        # Initialize resulting MAG
        dag = np.copy(cpdag_graph)

        # Turn circle component CPDAG^C into a DAG with no unshielded colliders.
        circle_cpdag = np.copy(cpdag_graph)
        # All lagged links are directed by time, remove them here
        circle_cpdag[:,:,1:] = ""
        # Also remove conflicting links
        circle_cpdag[circle_cpdag=="x-x"] = ""
        # Find undirected links, remove directed links
        for i, j, tau in zip(*np.where(circle_cpdag != "")):
            if circle_cpdag[i,j,0][1] == '?':
                raise ValueError("Invalid middle mark.")
            if circle_cpdag[i,j,0] == "-->":
                circle_cpdag[i,j,0] = ""

        # Iterate through simplicial nodes
        simplicial_node = self._get_simplicial_node(circle_cpdag,
                                                    variable_order)
        while simplicial_node is not None:

            # Choose such a vertex V1 and orient any edges incident to V1 into
            # V1 in the MAG And remove V1 and these edges from the circle
            # component PAG
            (j, adj_j) = simplicial_node
            for var in adj_j:
                dag[var, j, 0] = "-->"
                dag[j, var, 0] = "<--"
                circle_cpdag[var, j, 0] = circle_cpdag[j, var, 0] = "" 

            # Iterate
            simplicial_node = self._get_simplicial_node(circle_cpdag,
                                                    variable_order)

        return dag

    def convert_to_string_graph(self, graph_bool):
        """Converts the 0,1-based graph returned by PCMCI to a string array
        with links '-->'.

        Parameters
        ----------
        graph_bool : array
            0,1-based graph array output by PCMCI.

        Returns
        -------
        graph : array
            graph as string array with links '-->'.
        """

        graph = np.zeros(graph_bool.shape, dtype='<U3')
        graph[:] = ""
        # Lagged links
        graph[:,:,1:][graph_bool[:,:,1:]==1] = "-->"
        # Unoriented contemporaneous links
        graph[:,:,0][np.logical_and(graph_bool[:,:,0]==1, 
                                    graph_bool[:,:,0].T==1)] = "o-o"
        # Conflicting contemporaneous links
        graph[:,:,0][np.logical_and(graph_bool[:,:,0]==2, 
                                    graph_bool[:,:,0].T==2)] = "x-x"
        # Directed contemporaneous links
        for (i,j) in zip(*np.where(
            np.logical_and(graph_bool[:,:,0]==1, graph_bool[:,:,0].T==0))):
            graph[i,j,0] = "-->"
            graph[j,i,0] = "<--"

        return graph

    def symmetrize_p_and_val_matrix(self, p_matrix, val_matrix, link_assumptions, conf_matrix=None):
        """Symmetrizes the p_matrix, val_matrix, and conf_matrix based on link_assumptions
           and the larger p-value.

        Parameters
        ----------
        val_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of test statistic values.
        p_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of p-values. Set to 1 if val_only=True.
        conf_matrix : array of shape [N, N, tau_max+1,2]
            Estimated matrix of confidence intervals of test statistic values.
            Only computed if set in cond_ind_test, where also the percentiles
            are set.
        link_assumptions : dict or None
            Dictionary of form {j:{(i, -tau): link_type, ...}, ...} specifying
            assumptions about links. This initializes the graph with entries
            graph[i,j,tau] = link_type. For example, graph[i,j,0] = '-->'
            implies that a directed link from i to j at lag 0 must exist.
            Valid link types are 'o-o', '-->', '<--'. In addition, the middle
            mark can be '?' instead of '-'. Then '-?>' implies that this link
            may not exist, but if it exists, its orientation is '-->'. Link
            assumptions need to be consistent, i.e., graph[i,j,0] = '-->'
            requires graph[j,i,0] = '<--' and acyclicity must hold. If a link
            does not appear in the dictionary, it is assumed absent. That is,
            if link_assumptions is not None, then all links have to be specified
            or the links are assumed absent.
        Returns
        -------
        val_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of test statistic values.
        p_matrix : array of shape [N, N, tau_max+1]
            Estimated matrix of p-values. Set to 1 if val_only=True.
        conf_matrix : array of shape [N, N, tau_max+1,2]
            Estimated matrix of confidence intervals of test statistic values.
            Only computed if set in cond_ind_test, where also the percentiles
            are set.
        """

        # Symmetrize p_matrix and val_matrix and conf_matrix
        for i in range(self.N):
            for j in range(self.N):
                # If both the links are present in link_assumptions, symmetrize using maximum p-value
                # if ((i, 0) in link_assumptions[j] and (j, 0) in link_assumptions[i]):
                if (i, 0) in link_assumptions[j]:
                    if link_assumptions[j][(i, 0)] in ["o-o", 'o?o']:
                        if (p_matrix[i, j, 0]
                                >= p_matrix[j, i, 0]):
                            p_matrix[j, i, 0] = p_matrix[i, j, 0]
                            val_matrix[j, i, 0] = val_matrix[i, j, 0]
                            if conf_matrix is not None:
                                conf_matrix[j, i, 0] = conf_matrix[i, j, 0]

                    # If only one of the links is present in link_assumptions, symmetrize using the p-value of the link present
                    # elif ((i, 0) in link_assumptions[j] and (j, 0) not in link_assumptions[i]):
                    elif link_assumptions[j][(i, 0)] in ["-->", '-?>']:
                        p_matrix[j, i, 0] = p_matrix[i, j, 0]
                        val_matrix[j, i, 0] = val_matrix[i, j, 0]
                        if conf_matrix is not None:
                            conf_matrix[j, i, 0] = conf_matrix[i, j, 0]
                    else:
                        # Links not present in link_assumptions
                        pass

        # Return the values as a dictionary and store in class
        results = {'val_matrix': val_matrix,
                   'p_matrix': p_matrix,
                   'conf_matrix': conf_matrix}
        return results

    def run_sliding_window_of(self, method, method_args, 
                        window_step,
                        window_length,
                        conf_lev = 0.9,
                        ):
        """Runs chosen method on sliding windows taken from DataFrame.

        The function returns summary_results and all_results (containing the
        individual window results). summary_results contains val_matrix_mean
        and val_matrix_interval, the latter containing the confidence bounds for
        conf_lev. If the method also returns a graph, then 'most_frequent_links'
        containing the most frequent link outcome (either 0 or 1 or a specific
        link type) in each entry of graph, as well as 'link_frequency',
        containing the occurence frequency of the most frequent link outcome,
        are returned.

        Parameters
        ----------
        method : str
            Chosen method among valid functions in PCMCI.
        method_args : dict
            Arguments passed to method.
        window_step : int
            Time step of windows.
        window_length : int
            Length of sliding window.
        conf_lev : float, optional (default: 0.9)
            Two-sided confidence interval for summary results.

        Returns
        -------
        Dictionary of results for every sliding window.
        """

        valid_methods = ['run_pc_stable',
                          'run_mci',
                          'get_lagged_dependencies',
                          'run_fullci',
                          'run_bivci',
                          'run_pcmci',
                          'run_pcalg',
                          'run_lpcmci',
                          'run_jpcmciplus',
                          # 'run_pcalg_non_timeseries_data',
                          'run_pcmciplus',]

        if method not in valid_methods:
            raise ValueError("method must be one of %s" % str(valid_methods))

        if self.dataframe.reference_points_is_none is False:
            raise ValueError("Reference points are not accepted in "
                             "sliding windows analysis, align data before and use masking"
                             " and/or missing values.")

        T = self.dataframe.largest_time_step

        if self.cond_ind_test.recycle_residuals:
            # recycle_residuals clashes with sliding windows...
            raise ValueError("cond_ind_test.recycle_residuals must be False.")

        if self.verbosity > 0:
            print("\n##\n## Running sliding window analysis of %s " % method +
                  "\n##\n" +
                  "\nwindow_step = %s \n" % window_step +
                  "\nwindow_length = %s \n" % window_length
                  )

        original_reference_points = deepcopy(self.dataframe.reference_points)

        window_start_points = np.arange(0, T - window_length, window_step)
        n_windows = len(window_start_points)

        if len(window_start_points) == 0:
            raise ValueError("Empty list of windows, check window_length and window_step!")

        window_results = {}
        for iw, w in enumerate(window_start_points):
            if self.verbosity > 0:
                print("\n# Window start %s (%d/%d) \n" %(w, iw+1, len(window_start_points)))                
            # Construct reference_points from window
            time_window = np.arange(w, w + window_length, 1)
            # Remove points beyond T
            time_window = time_window[time_window < T]

            self.dataframe.reference_points = time_window
            window_res = deepcopy(getattr(self, method)(**method_args))

            # Aggregate val_matrix and other arrays to new arrays with
            # windows as first dimension. Lists and other objects
            # are stored in dictionary
            for key in window_res:
                res_item = window_res[key]
                if iw == 0:
                    if type(res_item) is np.ndarray:
                        window_results[key] = np.empty((n_windows,) 
                                                     + res_item.shape,
                                                     dtype=res_item.dtype) 
                    else:
                        window_results[key] = {}
                
                window_results[key][iw] = res_item

        # Reset to original_reference_points data for further analyses
        # self.dataframe.values[0] = original_data
        self.dataframe.reference_points = original_reference_points

        # Generate summary results
        summary_results = self.return_summary_results(results=window_results, 
                                                      conf_lev=conf_lev)

        return {'summary_results': summary_results, 
                'window_results': window_results}

    def run_bootstrap_of(self, method, method_args,
                        boot_samples=100,
                        boot_blocklength=1,
                        conf_lev=0.9, , aggregation="majority", seed=None):
        """Runs chosen method on bootstrap samples drawn from DataFrame.

        Bootstraps for tau=0 are drawn from [2xtau_max, ..., T] and all lagged
        variables constructed in DataFrame.construct_array are consistently
        shifted with respect to this bootstrap sample to ensure that lagged
        relations in the bootstrap sample are preserved.

        The function returns summary_results and all_results (containing the
        individual bootstrap results). summary_results contains
        val_matrix_mean and val_matrix_interval, the latter containing the
        confidence bounds for conf_lev. If the method also returns a graph,
        then 'most_frequent_links' containing the most frequent link outcome
        (specific link type) in each entry of graph, as well
        as 'link_frequency', containing the occurence frequency of the most
        frequent link outcome, are returned. Two aggregation methods are
        available for 'most_frequent_links'. By default, "majority"
        provides the most frequent link outcome. Alternatively 
        "no_edge_majority" provides an alternative aggregation strategy.
        As explained in Debeire et al. (2024), in the first step of this 
        alternative approach, the orientation of edges is ignored, and the 
        focus is only on determining the adjacency of each pair of vertices. 
        This is done through majority voting between no edge and all other 
        edge types. In the second step, the adjacencies identified in the
        first step are oriented based on majority voting. This alternative 
        approach ensures that no edge can only be voted on if it appears 
        in more than half of the bootstrap ensemble of graphs.

        Assumes that method uses cond_ind_test.run_test() function with cut_off
        = '2xtau_max'.

        Utilizes parallelization via joblib.

        Parameters
        ----------
        method : str
            Chosen method among valid functions in PCMCI.
        method_args : dict
            Arguments passed to method.
        boot_samples : int
            Number of bootstrap samples to draw.
        boot_blocklength : int, optional (default: 1)
            Block length for block-bootstrap.
        conf_lev : float, optional (default: 0.9)
            Two-sided confidence interval for summary results.
        seed : int, optional(default = None)
            Seed for RandomState (default_rng)
        aggregation : str, optional (default: "majority")
            Chosen aggregation strategy: "majority" or "no_edge_majority".

        Returns
        -------
        Dictionary of summary results and results for every bootstrap sample.
        """

        valid_methods = ['run_pc_stable',
                          'run_mci',
                          'get_lagged_dependencies',
                          'run_fullci',
                          'run_bivci',
                          'run_pcmci',
                          'run_pcalg',
                          'run_pcalg_non_timeseries_data',
                          'run_pcmciplus',
                          'run_lpcmci',
                          'run_jpcmciplus',
                          ]
        if method not in valid_methods:
            raise ValueError("method must be one of %s" % str(valid_methods))

        T = self.dataframe.largest_time_step
        seed_sequence = np.random.SeedSequence(seed)
        #global_random_state = np.random.default_rng(seed)

        # Extract tau_max to construct bootstrap draws
        if 'tau_max' not in method_args:
            raise ValueError("tau_max must be explicitely set in method_args.")
        tau_max = method_args['tau_max']

        if self.cond_ind_test.recycle_residuals:
            # recycle_residuals clashes with bootstrap draws...
            raise ValueError("cond_ind_test.recycle_residuals must be False.")

        if self.verbosity > 0:
            print("\n##\n## Running Bootstrap of %s " % method +
                  "\n##\n" +
                  "\nboot_samples = %s \n" % boot_samples +
                  "\nboot_blocklength = %s \n" % boot_blocklength
                  )

        # Set bootstrap attribute to be passed to dataframe
        self.dataframe.bootstrap = {}
        self.dataframe.bootstrap['boot_blocklength'] = boot_blocklength

        boot_results = {}
        #for b in range(boot_samples):
            # Generate random state for this boot and set it in dataframe
            # which will generate a draw with replacement
            #boot_seed = global_random_state.integers(0, boot_samples, 1)
            #boot_random_state = np.random.default_rng(boot_seed)
            #self.dataframe.bootstrap['random_state'] = boot_random_state

        child_seeds = seed_sequence.spawn(boot_samples)

        aggregated_results = Parallel(n_jobs=-1)(
            delayed(self.parallelized_bootstraps)(method, method_args, boot_seed=child_seeds[b]) for
            b in range(boot_samples))

        for b in range(boot_samples):
            # Aggregate val_matrix and other arrays to new arrays with
            # boot_samples as first dimension. Lists and other objects
            # are stored in dictionary
            boot_res = aggregated_results[b]
            for key in boot_res:
                res_item = boot_res[key]
                if type(res_item) is np.ndarray:
                    if b == 0:
                        boot_results[key] = np.empty((boot_samples,) 
                                                     + res_item.shape,
                                                     dtype=res_item.dtype) 
                    boot_results[key][b] = res_item
                else:
                    if b == 0:
                        boot_results[key] = {}
                    boot_results[key][b] = res_item

        # Generate summary results
        summary_results = self.return_summary_results(results=boot_results, 
                                                      conf_lev=conf_lev,
                                                      aggregation=aggregation)

        # Reset bootstrap to None
        self.dataframe.bootstrap = None

        return {'summary_results': summary_results, 
                'boot_results': boot_results}

    def parallelized_bootstraps(self, method, method_args, boot_seed):
        # Pass seed sequence for this boot and set it in dataframe
        # which will generate a draw with replacement
        boot_random_state = np.random.default_rng(boot_seed)
        self.dataframe.bootstrap['random_state'] = boot_random_state
        boot_res = getattr(self, method)(**method_args)
        return boot_res

    @staticmethod
    def return_summary_results(results, conf_lev=0.9, aggregation="majority"):
        """Return summary results for causal graphs.

        The function returns summary_results of an array of PCMCI(+) results.
        Summary_results contains val_matrix_mean and val_matrix_interval, the latter 
        containing the confidence bounds for conf_lev. If the method also returns a graph,
        then 'most_frequent_links' containing the most frequent link outcome 
        (either 0 or 1 or a specific link type) in each entry of graph, as well 
        as 'link_frequency', containing the occurence frequency of the most 
        frequent link outcome, are returned. Two aggregation methods are
        available for 'most_frequent_links'. By default, "majority"
        provides the most frequent link outcome. Alternatively 
        "no_edge_majority" provides an alternative aggregation strategy.
        As explained in Debeire et al. (2024), in the first step of this 
        alternative approach, the orientation of edges is ignored, and the 
        focus is only on determining the adjacency of each pair of vertices. 
        This is done through majority voting between no edge and all other 
        edge types. In the second step, the adjacencies identified in the
        first step are oriented based on majority voting. This alternative 
        approach ensures that no edge can only be voted on if it appears 
        in more than half of the bootstrap ensemble of graphs.

        Parameters
        ----------
        results : dict
            Results dictionary where the numpy arrays graph and val_matrix are
            of shape (n_results, N, N, tau_max + 1).
        conf_lev : float, optional (default: 0.9)
            Two-sided confidence interval for summary results.
        aggregation : str, optional (default: "majority")
            Chosen aggregation strategy: "majority" or "no_edge_majority".
        Returns
        -------
        Dictionary of summary results.
        """

        valid_aggregations = {"majority", "no_edge_majority"}
        if aggregation not in valid_aggregations:
            raise ValueError(f"Invalid aggregation mode: {aggregation}. Expected one of {valid_aggregations}")

        # Generate summary results
        summary_results = {}

        if 'graph' in results:
            n_results, N, N, tau_max_plusone = results['graph'].shape
            tau_max = tau_max_plusone - 1
            # print(repr(results['graph']))
            summary_results['most_frequent_links'] = np.zeros((N, N, tau_max_plusone),
                                dtype=results['graph'][0].dtype)
            summary_results['link_frequency'] = np.zeros((N, N, tau_max_plusone),
                                dtype='float')
            
            #preferred order in case of ties with the spirit of 
            #keeping the least assertive and most cautious claims in the presence of ties.
            #In case of ties between other link types, a conflicting link "x-x" is assigned
            preferred_order = [
            "",       # No link (most conservative)
            #"o?o",    # No claim made (lag 0 only)
            #"<?>",    # Neither is ancestor
            "x-x",    # Conflict (used to break <--> vs --> vs <-- ties)
            "o-o",    # Undirected link (lag 0 only)
            # "<-o",    # X^i not ancestor, but linked (lag 0 only)
            # "o->",    # X^j not ancestor, but linked
            # rest is solved by conflict
            # "<->",
            # "-->",
            # "<--",
            ]

            for (i, j) in itertools.product(range(N), range(N)):
                for abstau in range(0, tau_max + 1):
                    links, counts = np.unique(results['graph'][:,i,j,abstau], 
                                        return_counts=True)
                    list_of_most_freq = links[counts == counts.max()]
                    if aggregation=="majority":
                        if len(list_of_most_freq) == 1:
                            choice = list_of_most_freq[0]
                        else:
                            ordered_list = [link for link in preferred_order
                                            if link in list_of_most_freq]
                            if len(ordered_list) == 0:
                                choice = "x-x"
                            else:
                                choice = ordered_list[0]
                        summary_results['most_frequent_links'][i,j, abstau] = choice
                        summary_results['link_frequency'][i,j, abstau] = \
                                    counts[counts == counts.max()].sum()/float(n_results)

                    elif aggregation=="no_edge_majority":
                        if counts[links == ""].size == 0: #handle the case where there is no "" in links
                            freq_of_no_edge=0
                        else:
                            # make scalar count (counts[...] returns a 1-element array)
                            freq_of_no_edge = int(counts[links == ""].sum())
                            
                        freq_of_adjacency = n_results - freq_of_no_edge
                        if freq_of_adjacency > freq_of_no_edge:
                            adja_links = np.delete(links,np.where(links == ""))
                            adja_counts = np.delete(counts,np.where(links == ""))
                            list_of_most_freq_adja = adja_links[adja_counts == adja_counts.max()]
                            if len(list_of_most_freq_adja) == 1:
                                choice = list_of_most_freq_adja[0]
                            else:
                                ordered_list = [link for link in preferred_order
                                                if link in list_of_most_freq_adja]
                                if len(ordered_list) == 0:
                                    choice = "x-x"
                                else:
                                    choice = ordered_list[0]
                            summary_results['most_frequent_links'][i,j, abstau] = choice
                            summary_results['link_frequency'][i,j, abstau] = \
                                    adja_counts[adja_counts == adja_counts.max()].sum()/float(n_results)
                        else: 
                            choice= ""
                            summary_results['most_frequent_links'][i,j, abstau] = choice
                            summary_results['link_frequency'][i,j, abstau] = \
                                    freq_of_no_edge/float(n_results)
        # Confidence intervals for val_matrix; interval is two-sided
        c_int = (1. - (1. - conf_lev)/2.)
        summary_results['val_matrix_mean'] = np.mean(
                                    results['val_matrix'], axis=0)

        summary_results['val_matrix_interval'] = np.stack(np.percentile(
                                    results['val_matrix'], axis=0,
                                    q = [100*(1. - c_int), 100*c_int]), axis=3)
        return summary_results

    @staticmethod
    def graph_to_dict(graph):
        """Helper function to convert graph to dictionary of links.

        Parameters
        ---------
        graph : array of shape (N, N, tau_max+1)
            Matrix format of graph in string format.

        Returns
        -------
        links : dict
            Dictionary of form {0:{(0, -1): o-o, ...}, 1:{...}, ...}.
        """
        N = graph.shape[0]

        links = dict([(j, {}) for j in range(N)])

        for (i, j, tau) in zip(*np.where(graph!='')):
            links[j][(i, -tau)] = graph[i,j,tau]

        return links

    # @staticmethod
    def _dict_to_graph(self, links, tau_max=None):
        """Helper function to convert dictionary of links to graph.

        Parameters
        ---------
        links : dict
            Dictionary of form {0:{(0, -1): 'o-o'}, ...}, 1:{...}, ...}.

        Returns
        -------
        graph : array of shape (N, N, tau_max+1)
            Matrix format of graph in string format.
        """

        N = len(links)

        # Get maximum time lag
        max_lag = 0
        for j in range(N):
            for link in links[j]:
                var, lag = link
                if isinstance(links[j], dict):
                    link_type = links[j][link]
                    if link_type != "":
                        max_lag = max(max_lag, abs(lag))
                else:
                    max_lag = max(max_lag, abs(lag))

        if tau_max is None:
            tau_max = max_lag
        else:
            if tau_max < max_lag:
                raise ValueError("maxlag(links) > tau_max")

        graph = np.zeros((N, N, tau_max + 1), dtype='<U3')
        graph[:] = ""
        for j in range(N):
            for link in links[j]:
                i, tau = link
                if isinstance(links[j], dict):
                    link_type = links[j][link]
                    graph[i, j, abs(tau)] = link_type
                else:
                    graph[i, j, abs(tau)] = '-->'

        return graph

    @staticmethod
    def get_graph_from_dict(links, tau_max=None):
        """Helper function to convert dictionary of links to graph array format.

        Parameters
        ---------
        links : dict
            Dictionary of form {0:[((0, -1), coeff, func), ...], 1:[...], ...}.
            Also format {0:[(0, -1), ...], 1:[...], ...} is allowed.
        tau_max : int or None
            Maximum lag. If None, the maximum lag in links is used.

        Returns
        -------
        graph : array of shape (N, N, tau_max+1)
            Matrix format of graph with 1 for true links and 0 else.
        """

        def _get_minmax_lag(links):
            """Helper function to retrieve tau_min and tau_max from links.
            """

            N = len(links)

            # Get maximum time lag
            min_lag = np.inf
            max_lag = 0
            for j in range(N):
                for link_props in links[j]:
                    if len(link_props) > 2:
                        var, lag = link_props[0]
                        coeff = link_props[1]
                        # func = link_props[2]
                        if coeff != 0.:
                            min_lag = min(min_lag, abs(lag))
                            max_lag = max(max_lag, abs(lag))
                    else:
                        var, lag = link_props
                        min_lag = min(min_lag, abs(lag))
                        max_lag = max(max_lag, abs(lag))   

            return min_lag, max_lag

        N = len(links)

        # Get maximum time lag
        min_lag, max_lag = _get_minmax_lag(links)

        # Set maximum lag
        if tau_max is None:
            tau_max = max_lag
        else:
            if max_lag > tau_max:
                raise ValueError("tau_max is smaller than maximum lag = %d "
                                 "found in links, use tau_max=None or larger "
                                 "value" % max_lag)

        graph = np.zeros((N, N, tau_max + 1), dtype='<U3')
        for j in links.keys():
            for link_props in links[j]:
                if len(link_props) > 2:
                    var, lag = link_props[0]
                    coeff = link_props[1]
                    if coeff != 0.:
                        graph[var, j, abs(lag)] = "-->"
                        if lag == 0:
                            graph[j, var, 0] = "<--"
                else:
                    var, lag = link_props
                    graph[var, j, abs(lag)] = "-->"
                    if lag == 0:
                        graph[j, var, 0] = "<--"

        return graph

    @staticmethod
    def build_link_assumptions(link_assumptions_absent_link_means_no_knowledge,
                               n_component_time_series,
                               tau_max,
                               tau_min=0):
        
        out = {j: {(i, -tau_i): ("o?>" if tau_i > 0 else "o?o")
             for i in range(n_component_time_series) for tau_i in range(tau_min, tau_max+1)
             if (tau_i > 0 or i != j)} for j in range(n_component_time_series)}
        
        for j, links_j in link_assumptions_absent_link_means_no_knowledge.items():
            for (i, lag_i), link_ij in links_j.items():
                if link_ij == "": 
                    del out[j][(i, lag_i)]
                else:
                    out[j][(i, lag_i)] = link_ij
        return out
